"""Mixin class for MOM and MOM6 drivers

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
:license: Apache License, Version 2.0, see LICENSE for details
"""

import os

import cftime


class MomMixin:

    def get_restart_datetime(self, restart_path):
        """Given a restart path, parse the restart files and
        return a cftime datetime (for date-based restart pruning)"""
        # Check for ocean_solo.res file
        ocean_solo_path = os.path.join(restart_path, 'ocean_solo.res')
        if not os.path.exists(ocean_solo_path):
            raise FileNotFoundError(
                'Cannot find ocean_solo.res file, which is required for '
                'date-based restart pruning')

        with open(ocean_solo_path, 'r') as ocean_solo:
            lines = ocean_solo.readlines()

        calendar_int = int(lines[0].split()[0])
        cftime_calendars = {
            1: "360_day",
            2: "julian",
            3: "proleptic_gregorian",
            4: "noleap"
        }
        calendar = cftime_calendars[calendar_int]

        last_date_line = lines[-1].split()
        date_values = [int(i) for i in last_date_line[:6]]
        year, month, day, hour, minute, second = date_values
        return cftime.datetime(year, month, day, hour, minute, second,
                               calendar=calendar)



import os
import sys

from payu import envmod
from payu.profilers.profiler import Profiler


class OpenSpeedShop(Profiler):

    def setup(self):
        os.environ['OPENSS_RAWDATA_DIR'] = self.expt.work_path
        os.environ['OPENSS_DB_DIR'] = self.expt.work_path

    def load_modules(self):
        envmod.module('load', 'openspeedshop')

    def wrapper(self, cmd):
        # TODO: Get this from the "profilers" entry, not a separate one
        oss = self.expt.config.get('openspeedshop')
        if oss:
            oss_runcmd = oss.get('runcmd')
            if not oss_runcmd:
                print('payu: error: OpenSpeedShop requires an executable.')
                sys.exit(1)

            cmd = '{0} "{1}"'.format(oss_runcmd, cmd)

            if oss_runcmd.startswith('osshwc'):
                oss_hwc = oss.get('hwc')
                if not oss_hwc:
                    print('payu: error: This OSS command requires hardware '
                          'counters.')
                    sys.exit(1)
                else:
                    cmd = ' '.join([cmd, oss_hwc])

        return cmd

    def postprocess(self):
        pass



import copy
import shutil
from pathlib import Path

import pytest
import git
from ruamel.yaml import YAML
from unittest.mock import patch

from payu.branch import add_restart_to_config, check_restart, switch_symlink
from payu.branch import checkout_branch, clone, list_branches
from payu.metadata import MetadataWarning
from payu.fsops import read_config

from test.common import cd
from test.common import tmpdir, ctrldir, labdir, archive_dir
from test.common import ctrldir_basename
from test.common import config as config_orig, write_config
from test.common import config_path, metadata_path
from test.common import make_expt_archive_dir


# Global config - Remove set experiment and metadata config
config = copy.deepcopy(config_orig)
config.pop("experiment")
config.pop("metadata")


@pytest.fixture(autouse=True)
def setup_and_teardown():
    # Create tmp, lab and control directories
    try:
        tmpdir.mkdir()
        labdir.mkdir()
        ctrldir.mkdir()
    except Exception as e:
        print(e)

    yield

    # Remove tmp directory
    try:
        shutil.rmtree(tmpdir)
    except Exception as e:
        print(e)


def setup_control_repository(path=ctrldir, set_config=True):
    """Return an new control repository"""
    if set_config:
        write_config(config, path=(path / "config.yaml"))
    else:
        (path / "newFile").touch()
    # Initialise a control repo
    repo = git.Repo.init(path)
    repo.index.add("*")
    # Commit the changes
    repo.index.commit("First commit - initialising repository")
    return repo


@pytest.mark.parametrize(
    "test_config, expected_config",
    [
        (
            """sync:
  restart: true
# Test comment
restart: old/path/to/restart
anotherField: 1
""",
            """sync:
  restart: true
# Test comment
restart: {0}
anotherField: 1
"""
        ),
        (
            """# Test comment

anotherField: 1""",
            """# Test comment

anotherField: 1
restart: {0}
"""
        ),
    ]
)
def test_add_restart_to_config(test_config, expected_config):
    """Test adding restart: path/to/restart to configuration file"""
    restart_path = tmpdir / "archive" / "tmpRestart"
    restart_path.mkdir(parents=True)

    expected_config = expected_config.format(restart_path)

    with config_path.open("w") as file:
        file.write(test_config)

    # Function to test
    with cd(ctrldir):
        add_restart_to_config(restart_path, config_path)

    with config_path.open("r") as file:
        updated_config = file.read()

    # Test order, comments are preserved
    assert updated_config == expected_config


def test_check_restart_relative_path():
    """Test an relative restart path is resolved to an absolute path"""
    restart_path = tmpdir / "archive" / "tmpRestart"
    restart_path.mkdir(parents=True)

    with cd(tmpdir):
        relative_restart_path = Path("archive") / "tmpRestart"
        assert not relative_restart_path.is_absolute()

        resolved_path = check_restart(relative_restart_path)
        assert resolved_path.is_absolute()


def test_check_restart_with_non_existent_restart():
    """Test restart path that does not exist raises a warning"""
    restart_path = tmpdir / "restartDNE"

    expected_msg = (f"Given restart path {restart_path} does not exist. "
                    f"Skipping setting 'restart' in config file")

    with cd(ctrldir):
        with pytest.warns(UserWarning, match=expected_msg):
            restart_path = check_restart(restart_path, labdir / "archive")

    assert restart_path is None


def test_check_restart_with_pre_existing_restarts_in_archive():
    """Test pre-existing restarts in archive raises a warning"""
    # Create pre-existing restart in archive
    archive_path = labdir / "archive"
    (archive_path / "restart000").mkdir(parents=True)

    # Create restart path in different archive
    restart_path = labdir / "different_archive" / "restart000"
    restart_path.mkdir(parents=True)

    expected_msg = (
        f"Pre-existing restarts found in archive: {archive_path}."
        f"Skipping adding 'restart: {restart_path}' to config file"
    )

    with cd(ctrldir):
        with pytest.warns(UserWarning, match=expected_msg):
            restart_path = check_restart(restart_path, archive_path)

    assert restart_path is None


def test_switch_symlink_when_symlink_and_archive_exists():
    # Pre-existing experiment symlink
    lab_archive = labdir / "archive"
    previous_archive_dir = lab_archive / "Experiment0"
    previous_archive_dir.mkdir(parents=True)

    archive_symlink = ctrldir / "archive"
    archive_symlink.symlink_to(previous_archive_dir)

    # New Experiment - Existing archive
    experiment_name = "Experiment1"
    archive_dir = lab_archive / experiment_name
    archive_dir.mkdir(parents=True)

    # Test Function
    switch_symlink(lab_archive, ctrldir, experiment_name, "archive")

    # Assert new symlink is created
    assert archive_symlink.exists() and archive_symlink.is_symlink()
    assert archive_symlink.resolve() == archive_dir


def test_switch_symlink_when_symlink_exists_but_no_archive():
    # Pre-existing experiment symlink
    lab_archive = labdir / "archive"
    previous_archive_dir = lab_archive / "Experiment0"
    previous_archive_dir.mkdir(parents=True)

    archive_symlink = ctrldir / "archive"
    archive_symlink.symlink_to(previous_archive_dir)

    # New Experiment
    experiment_name = "Experiment1"

    # Test Function
    switch_symlink(lab_archive, ctrldir, experiment_name, "archive")

    # Assert no symlink is created but previous one is removed
    assert not archive_symlink.exists()
    assert not archive_symlink.is_symlink()


def test_switch_symlink_when_no_symlink_exists_and_no_archive():
    # New Experiment
    experiment_name = "Experiment1"
    lab_archive = labdir / "archive"

    archive_symlink = ctrldir / "archive"

    # Test Function
    switch_symlink(lab_archive, ctrldir, experiment_name, "archive")

    # Assert no symlink
    assert not archive_symlink.exists()
    assert not archive_symlink.is_symlink()


def test_switch_symkink_when_previous_symlink_dne():
    # Point archive symlink to a directory that does not exist anymore
    lab_archive = labdir / "archive"
    previous_archive_dir = lab_archive / "ExperimentDNE"

    archive_symlink = ctrldir / "archive"
    archive_symlink.symlink_to(previous_archive_dir)

    # New Experiment
    experiment_name = "Experiment1"
    archive_dir = lab_archive / experiment_name
    archive_dir.mkdir(parents=True)

    # Test Function
    switch_symlink(lab_archive, ctrldir, experiment_name, "archive")

    # Assert new symlink is created
    assert archive_symlink.exists() and archive_symlink.is_symlink()
    assert archive_symlink.resolve() == archive_dir


def check_metadata(expected_uuid,
                   expected_experiment,
                   expected_parent_uuid=None,
                   metadata_file=metadata_path):
    """Helper function to read metadata file and assert changed as expected"""
    assert metadata_file.exists()
    metadata = YAML().load(metadata_file)
    assert metadata.get("experiment_uuid", None) == expected_uuid
    assert metadata.get("parent_experiment", None) == expected_parent_uuid

    # Assert archive exists for experiment name
    assert (archive_dir / expected_experiment / "metadata.yaml").exists()
    copied_metadata = YAML().load(metadata_file)
    assert copied_metadata == metadata


def check_branch_metadata(repo,
                          expected_current_branch,
                          expected_uuid,
                          expected_experiment,
                          expected_parent_uuid=None,
                          metadata_file=metadata_path):
    """Helper function for checking expected  branch and metadata"""
    # Check metadata
    check_metadata(expected_uuid,
                   expected_experiment,
                   expected_parent_uuid,
                   metadata_file=metadata_file)

    # Check cuurent branch
    assert str(repo.active_branch) == expected_current_branch

    # Check last commit message
    expected_commit_msg = f"Updated metadata. Experiment UUID: {expected_uuid}"
    assert repo.head.commit.message == expected_commit_msg


@patch("uuid.uuid4")
def test_checkout_branch(mock_uuid):
    repo = setup_control_repository()

    # Mock uuid1 value
    uuid1 = "8ddc1985-b7d0-4d4d-884f-061ecd90d478"
    mock_uuid.return_value = uuid1

    with cd(ctrldir):
        # Test checkout new branch (with no existing metadata)
        checkout_branch(branch_name="Branch1",
                        is_new_branch=True,
                        lab_path=labdir)

    # Check current branch, new commit was added, and metadata created
    branch1_experiment_name = f"{ctrldir_basename}-Branch1-8ddc1985"
    check_branch_metadata(repo,
                          expected_uuid=uuid1,
                          expected_current_branch="Branch1",
                          expected_experiment=branch1_experiment_name)

    # Save commit hash to check later on
    branch_1_commit_hash = repo.active_branch.object.hexsha

    # Mock uuid2 value
    uuid2 = "2de5b001-df08-4c0b-ab15-f47f8ad72929"
    mock_uuid.return_value = uuid2

    with cd(ctrldir):
        # Test checkout new branch from branch with existing metadata
        checkout_branch(branch_name="Branch2",
                        is_new_branch=True,
                        lab_path=labdir)

    # Check current branch, new commit was added, and metadata created
    branch2_experiment_name = f"{ctrldir_basename}-Branch2-2de5b001"
    check_branch_metadata(repo,
                          expected_uuid=uuid2,
                          expected_current_branch="Branch2",
                          expected_experiment=branch2_experiment_name)

    # Mock uuid3 value
    uuid3 = "98c99f06-260e-42cc-a23f-f113fae825e5"
    mock_uuid.return_value = uuid3

    with cd(ctrldir):
        # Test checkout new branch from starting branch with existing metadata
        checkout_branch(branch_name="Branch3",
                        is_new_branch=True,
                        start_point="Branch1",
                        lab_path=labdir)

    # Check current branch, new commit was added, and metadata created
    branch3_experiment_name = f"{ctrldir_basename}-Branch3-98c99f06"
    check_branch_metadata(repo,
                          expected_uuid=uuid3,
                          expected_current_branch="Branch3",
                          expected_experiment=branch3_experiment_name)

    # Check second to last commit was last commit on branch 1
    second_latest_commit = list(repo.iter_commits(max_count=2))[1]
    assert second_latest_commit.hexsha == branch_1_commit_hash

    with cd(ctrldir):
        # Test checkout existing branch with existing metadata
        checkout_branch(branch_name="Branch1",
                        lab_path=labdir)

    # Check metadata and commit has not changed on Branch1
    assert str(repo.active_branch) == "Branch1"
    check_metadata(expected_experiment=branch1_experiment_name,
                   expected_uuid=uuid1)

    # Assert commit hash is the same
    assert repo.active_branch.object.hexsha == branch_1_commit_hash


@patch("uuid.uuid4")
def test_checkout_existing_branch_with_no_metadata(mock_uuid):
    repo = setup_control_repository()

    # Create new branch
    repo.create_head("Branch1")

    # Mock uuid1 value
    uuid1 = "574ea2c9-2379-4484-86b4-1d1a0f820773"
    mock_uuid.return_value = uuid1
    expected_no_uuid_msg = (
        "No experiment uuid found in metadata. Generating a new uuid"
    )
    expected_no_archive_msg = (
        "No pre-existing archive found. Generating a new uuid"
    )

    with cd(ctrldir):
        # Test checkout existing branch with no existing metadata
        with pytest.warns(MetadataWarning, match=expected_no_uuid_msg):
            with pytest.warns(MetadataWarning, match=expected_no_archive_msg):
                checkout_branch(branch_name="Branch1",
                                lab_path=labdir)

    # Check metadata was created and commited
    branch1_experiment_name = f"{ctrldir_basename}-Branch1-574ea2c9"
    check_branch_metadata(repo,
                          expected_uuid=uuid1,
                          expected_current_branch="Branch1",
                          expected_experiment=branch1_experiment_name)


@patch("uuid.uuid4")
def test_checkout_branch_with_no_metadata_and_with_legacy_archive(mock_uuid):
    # Make experiment archive - This function creates legacy experiment archive
    make_expt_archive_dir(type="restart", index=0)

    # Setup repo
    repo = setup_control_repository()

    # Create new branch using git
    repo.create_head("Branch1")

    # Mock uuid1 value
    uuid1 = "df050eaf-c8bb-4b10-9998-e0202a1eabd2"
    mock_uuid.return_value = uuid1
    expected_no_uuid_msg = (
        "No experiment uuid found in metadata. Generating a new uuid"
    )

    with cd(ctrldir):
        # Test checkout existing branch (with no existing metadata)
        # and with pre-existing archive
        with pytest.warns(MetadataWarning, match=expected_no_uuid_msg):
            checkout_branch(branch_name="Branch1",
                            lab_path=labdir)

    # Check metadata was created and commited - with legacy experiment name
    branch1_experiment_name = ctrldir_basename
    check_branch_metadata(repo,
                          expected_uuid=uuid1,
                          expected_current_branch="Branch1",
                          expected_experiment=branch1_experiment_name)


@patch("uuid.uuid4")
def test_checkout_new_branch_existing_legacy_archive(mock_uuid):
    # Using payu checkout new branch should generate new uuid,
    # and experiment name - even if there"s a legacy archive
    repo = setup_control_repository()

    # Add archive under legacy name
    restart_path = Path(make_expt_archive_dir(type="restart"))

    # Mock uuid1 value
    uuid1 = "d4437aae-8370-4567-a698-94d00ba87cdc"
    mock_uuid.return_value = uuid1

    with cd(ctrldir):
        # Test checkout new branch (with no existing metadata)
        checkout_branch(branch_name="Branch1",
                        is_new_branch=True,
                        restart_path=restart_path,
                        config_path=config_path,
                        lab_path=labdir)

    # Check metadata was created and commited - with branch-uuid aware name
    branch1_experiment_name = f"{ctrldir_basename}-Branch1-d4437aae"
    check_branch_metadata(repo,
                          expected_uuid=uuid1,
                          expected_current_branch="Branch1",
                          expected_experiment=branch1_experiment_name)

    # Check restart path was added to configuration file
    config = read_config(config_path)
    assert config["restart"] == str(restart_path)


def test_checkout_branch_with_no_config():
    # Initialise a control repo with no config
    repo = setup_control_repository(set_config=False)

    repo.create_head("Branch1")

    with cd(ctrldir):
        # Test checkout branch that has no config raise error
        with pytest.raises(FileNotFoundError):
            checkout_branch(branch_name="Branch1",
                            lab_path=labdir)

    assert not metadata_path.exists()


@patch("uuid.uuid4")
def test_checkout_branch_with_restart_path(mock_uuid):
    # Make experiment archive restart - starting with no metadata
    restart_path = tmpdir / "remote_archive" / "restart0123"
    restart_path.mkdir(parents=True)

    # Setup repo
    repo = setup_control_repository()

    # Mock uuid1 value
    uuid1 = "df050eaf-c8bb-4b10-9998-e0202a1eabd2"
    mock_uuid.return_value = uuid1

    with cd(ctrldir):
        # Test checkout with restart path with no metadata
        checkout_branch(is_new_branch=True,
                        branch_name="Branch1",
                        lab_path=labdir,
                        restart_path=restart_path)

    # Check metadata
    experiment1_name = f"{ctrldir_basename}-Branch1-df050eaf"
    check_branch_metadata(repo,
                          expected_current_branch='Branch1',
                          expected_uuid=uuid1,
                          expected_experiment=experiment1_name)

    # Create restart directory in Branch1 archive
    restart_path = archive_dir / experiment1_name / 'restart0123'
    restart_path.mkdir()

    # Mock uuid2 value
    uuid2 = "9cc04c9b-f13d-4f1d-8a35-87146a4381ef"
    mock_uuid.return_value = uuid2

    with cd(ctrldir):
        # Test checkout with restart path with metadata
        checkout_branch(is_new_branch=True,
                        branch_name="Branch2",
                        lab_path=labdir,
                        restart_path=restart_path)

    # Check metadta - Check parent experiment is experment 1 UUID
    experiment2_name = f"{ctrldir_basename}-Branch2-9cc04c9b"
    check_branch_metadata(repo,
                          expected_current_branch='Branch2',
                          expected_uuid=uuid2,
                          expected_experiment=experiment2_name,
                          expected_parent_uuid=uuid1)


@patch("payu.laboratory.Laboratory.initialize")
def test_checkout_laboratory_path_error(mock_lab_initialise):
    mock_lab_initialise.side_effect = PermissionError

    repo = setup_control_repository()
    current_commit = repo.active_branch.object.hexsha

    with cd(ctrldir):
        # Test raises a permission error
        with pytest.raises(PermissionError):
            checkout_branch(branch_name="Branch1",
                            is_new_branch=True,
                            lab_path=labdir)

    # Assert new commit has not been added
    assert repo.active_branch.object.hexsha == current_commit

    assert str(repo.active_branch) == "Branch1"
    assert not (ctrldir / "metadata.yaml").exists()

    # Test removing lab directory
    shutil.rmtree(labdir)
    mock_lab_initialise.side_effect = None
    with cd(ctrldir):
        # Test runs without an error - directory is initialised
        checkout_branch(branch_name="Branch2",
                        is_new_branch=True,
                        lab_path=labdir)

    # Assert new commit has been added
    assert repo.active_branch.object.hexsha != current_commit


@patch("uuid.uuid4")
def test_clone(mock_uuid):
    # Create a repo to clone
    source_repo_path = tmpdir / "sourceRepo"
    source_repo_path.mkdir()
    source_repo = setup_control_repository(path=source_repo_path)
    source_main_branch = str(source_repo.active_branch)

    # Create and checkout branch
    branch1 = source_repo.create_head("Branch1")
    branch1.checkout()

    # Mock uuid1 value
    uuid1 = "9cc04c9b-f13d-4f1d-8a35-87146a4381ef"
    mock_uuid.return_value = uuid1

    # Test clone
    cloned_repo_path = tmpdir / "clonedRepo"
    clone(str(source_repo_path), cloned_repo_path, lab_path=labdir)

    # Check new commit added and expected metadata
    cloned_repo = git.Repo(cloned_repo_path)
    metadata_file = cloned_repo_path / "metadata.yaml"
    check_branch_metadata(repo=cloned_repo,
                          expected_current_branch="Branch1",
                          expected_uuid=uuid1,
                          expected_experiment="clonedRepo-Branch1-9cc04c9b",
                          metadata_file=metadata_file)
    branch_1_commit_hash = cloned_repo.active_branch.object.hexsha

    cloned_repo.git.checkout(source_main_branch)

    # Test clone of a clone - adding a new branch
    uuid2 = "fd7b4804-d306-4a18-9d95-a8f565abfc9a"
    mock_uuid.return_value = uuid2

    # Run clone
    with cd(tmpdir):
        clone(str(cloned_repo_path), Path("clonedRepo2"),
              lab_path=labdir, new_branch_name="Branch2", branch="Branch1",
              parent_experiment=uuid1)

    # Check new commit added and expected metadata
    cloned_repo2 = git.Repo(tmpdir / "clonedRepo2")
    metadata_file = tmpdir / "clonedRepo2" / "metadata.yaml"
    check_branch_metadata(repo=cloned_repo2,
                          expected_current_branch="Branch2",
                          expected_uuid=uuid2,
                          expected_experiment="clonedRepo2-Branch2-fd7b4804",
                          expected_parent_uuid=uuid1,
                          metadata_file=metadata_file)

    # Check branched from Branch1
    second_latest_commit = list(cloned_repo2.iter_commits(max_count=2))[1]
    assert second_latest_commit.hexsha == branch_1_commit_hash

    # Check local branches
    assert [head.name for head in cloned_repo2.heads] == ["Branch1", "Branch2"]


@pytest.mark.parametrize(
    "start_point_type", ["commit", "tag"]
)
def test_clone_startpoint(start_point_type):
    # Create a repo to clone
    source_repo_path = tmpdir / "sourceRepo"
    source_repo_path.mkdir()
    source_repo = setup_control_repository(path=source_repo_path)

    # Create branch1
    branch1 = source_repo.create_head("Branch1")
    branch1_commit = branch1.object.hexsha
    if start_point_type == "tag":
        source_repo.create_tag('v1.0', ref=branch1.commit)
        start_point = 'v1.0'
    elif start_point_type == "commit":
        start_point = branch1_commit

    # Add another commit on main branch so the commit is different to branch1
    (source_repo_path / "mock_file.txt").touch()
    source_repo.index.add("mock_file.txt")
    source_repo.index.commit("Another commit with a mock file")

    source_repo_commit = source_repo.active_branch.object.hexsha
    assert source_repo_commit != branch1_commit

    # Run Clone
    cloned_repo_path = tmpdir / "clonedRepo"
    with cd(tmpdir):
        clone(
            repository=str(source_repo_path),
            directory=cloned_repo_path,
            lab_path=labdir,
            new_branch_name="Branch3",
            start_point=start_point
        )

    cloned_repo = git.Repo(cloned_repo_path)

    # Check branched starting from start point
    second_latest_commit = list(cloned_repo.iter_commits(max_count=2))[1]
    assert second_latest_commit.hexsha == branch1_commit

    # Latest commit is different (new commit from metadata)
    assert source_repo_commit != cloned_repo.active_branch.object.hexsha


def test_clone_with_relative_restart_path():
    """Test clone with a restart path that is relative with respect to
    the directory in which the clone command is run from"""
    # Create a repo to clone
    source_repo_path = tmpdir / "sourceRepo"
    source_repo_path.mkdir()
    setup_control_repository(path=source_repo_path)

    # Create restart path
    restart_path = tmpdir / "archive" / "tmpRestart"
    restart_path.mkdir(parents=True)
    relative_restart_path = Path("archive") / "tmpRestart"

    cloned_repo_path = tmpdir / "clonedRepo"
    with cd(tmpdir):
        # Run clone
        clone(repository=str(source_repo_path),
              directory=cloned_repo_path,
              lab_path=labdir,
              restart_path=relative_restart_path)

    # Test restart was added to config.yaml file
    with cd(cloned_repo_path):
        config = read_config()

    assert config["restart"] == str(restart_path)


def add_and_commit_metadata(repo, metadata):
    """Helper function to create/update metadata file and commit"""
    metadata_path = ctrldir / "metadata.yaml"
    YAML().dump(metadata, metadata_path)
    repo.index.add("*")
    repo.index.commit("Updated metadata.yaml")


def test_list_branches(capsys):
    # Create repo and a few branches with and without metadata files
    repo = setup_control_repository(set_config=False)
    # Leave main branch with no metadata file
    main_branch_name = str(repo.active_branch)

    # Branch 1 - has config but no metadata
    branch1 = repo.create_head("Branch1")
    branch1.checkout()
    write_config(config)
    repo.index.add("*")
    repo.index.commit("Added config.yaml")

    # Checkout and add metadata to new branch
    branch2 = repo.create_head("Branch2")
    branch2.checkout()
    write_config(config)
    branch_2_metadata = {
        "experiment_uuid": "b12345678",
    }
    add_and_commit_metadata(repo, branch_2_metadata)

    # New branch with no uuid in metadata
    branch3 = repo.create_head("Branch3")
    branch3.checkout()
    branch_3_metadata = {
        "email": "test@email.com",
        "contact": "TestUser"
    }
    add_and_commit_metadata(repo, branch_3_metadata)

    # Test list branches
    with cd(ctrldir):
        list_branches()

    expected_printed_output = f"""* Current Branch: Branch3
    No UUID in metadata file
Branch: Branch1
    No metadata file found
Branch: Branch2
    experiment_uuid: b12345678
Branch: {main_branch_name}
    No config file found"""
    captured = capsys.readouterr()
    assert captured.out.strip() == expected_printed_output

    # Test list branches with verbose set
    with cd(ctrldir):
        list_branches(verbose=True)

    expected_verbose_output = f"""* Current Branch: Branch3
    email: test@email.com
    contact: TestUser
Branch: Branch1
    No metadata file found
Branch: Branch2
    experiment_uuid: b12345678
Branch: {main_branch_name}
    No config file found"""
    captured = capsys.readouterr()
    assert captured.out.strip() == expected_verbose_output

    # Test remote branches
    cloned_repo_path = tmpdir / "cloned_repo"
    repo.clone(cloned_repo_path)

    with cd(cloned_repo_path):
        list_branches(remote=True)
    expected_remote_output = f"""* Current Branch: Branch3
    No UUID in metadata file
Remote Branch: Branch1
    No metadata file found
Remote Branch: Branch2
    experiment_uuid: b12345678
Remote Branch: Branch3
    No UUID in metadata file
Remote Branch: HEAD
    No UUID in metadata file
Remote Branch: {main_branch_name}
    No config file found"""
    captured = capsys.readouterr()
    assert captured.out.strip() == expected_remote_output



# coding: utf-8

# Local
from payu.laboratory import Laboratory
import payu.subcommands.args as args

# Configuration
title = 'init'
parameters = {'description': 'Initialize the model laboratory'}

arguments = [args.model, args.config, args.laboratory]


def runcmd(model_type, config_path, lab_path):

    lab = Laboratory(model_type, config_path, lab_path)
    lab.initialize()



"""Simple wrappers around git commands

Using the GitPython library for interacting with Git
"""

import warnings
from pathlib import Path
from typing import Optional, Union, List, Dict

import git
import configparser


class PayuBranchError(Exception):
    """Custom exception for payu branch operations"""


class PayuGitWarning(Warning):
    """Custom warning class - useful for testing"""


def get_git_repository(repo_path: Union[Path, str],
                       initialise: bool = False,
                       catch_error: bool = False) -> Optional[git.Repo]:
    """Return a PythonGit repository object at given path. If initialise is
    true, it will attempt to initialise a repository if it does not exist.
    Otherwise, if catch_error is true, it will return None"""
    try:
        repo = git.Repo(repo_path)
        return repo
    except git.exc.InvalidGitRepositoryError:
        if initialise:
            repo = git.Repo.init(repo_path)
            print(f"Initialised new git repository at: {repo_path}")
            return repo

        warnings.warn(
            f"Path is not a valid git repository: {repo_path}",
            PayuGitWarning
        )
        if catch_error:
            return None
        raise


class GitRepository:
    """Simple wrapper around git python's repo and methods"""

    def __init__(self,
                 repo_path: Union[Path, str],
                 repo: Optional[git.Repo] = None,
                 catch_error: bool = False):
        self.repo_path = repo_path

        # Initialise git repository object
        if repo is None:
            repo = get_git_repository(repo_path, catch_error=catch_error)
        self.repo = repo

    def get_branch_name(self) -> Optional[str]:
        """Return the current git branch or None if repository path is
        not a git repository"""
        if self.repo:
            return str(self.repo.active_branch)

    def get_hash(self) -> Optional[str]:
        """Return the current git commit hash or None if repository path is
          not a git repository"""
        if self.repo:
            return self.repo.active_branch.object.hexsha

    def get_origin_url(self) -> Optional[str]:
        """Return url of remote origin if it exists"""
        if self.repo and self.repo.remotes and self.repo.remotes.origin:
            return self.repo.remotes.origin.url

    def get_user_info(self, config_key: str) -> Optional[str]:
        """Return git config user info, None otherwise. Used for retrieving
        name and email saved in git"""
        if self.repo is None:
            return

        try:
            config_reader = self.repo.config_reader()
            return config_reader.get_value('user', config_key)
        except (configparser.NoSectionError, configparser.NoOptionError):
            # No git config set for user.$config_key
            return

    def commit(self,
               commit_message: str,
               paths_to_commit: List[Union[Path, str]]) -> None:
        """Add a git commit of changes to paths"""
        if self.repo is None:
            return

        # Un-stage any pre-existing changes
        self.repo.index.reset()

        # Check if paths to commit have changed, or it is an untracked file
        changes = False
        untracked_files = [Path(self.repo_path) / path
                           for path in self.repo.untracked_files]
        for path in paths_to_commit:
            if self.repo.git.diff(None, path) or path in untracked_files:
                self.repo.index.add([path])
                changes = True

        # Run commit if there's changes
        if changes:
            self.repo.index.commit(commit_message)
            print(commit_message)

    def local_branches_dict(self) -> Dict[str, git.Head]:
        """Return a dictionary mapping local branch names to git.Head
        objects"""
        branch_names_dict = {}
        for head in self.repo.heads:
            branch_names_dict[head.name] = head
        return branch_names_dict

    def remote_branches_dict(self) -> Dict[str, git.Head]:
        """Return a dictionary mapping remote branch names to git.Head
        objects"""
        branch_names_dict = {}
        for remote in self.repo.remotes:
            try:
                remote.fetch()
            except git.exc.GitCommandError:
                warnings.warn(
                    f"Failed to fetch from remote repository: {remote.name} " +
                    f"(url: {remote.url}). Payu is not able to determine " +
                    "remote branch names, which are used in payu checkout " +
                    "to check if a branch already exists, or when creating " +
                    "a new branch from a remote branch.",
                    PayuGitWarning
                )
                continue

            for ref in remote.refs:
                branch_names_dict[ref.remote_head] = ref
        return branch_names_dict

    def checkout_branch(self,
                        branch_name: str,
                        new_branch: bool = False,
                        start_point: Optional[str] = None) -> None:
        """Checkout branch and create branch if specified"""
        # First check for staged changes
        if self.repo.is_dirty(index=True, working_tree=False):
            raise PayuBranchError(
                "There are staged git changes. Please stash or commit them "
                "before running the checkout command again.\n"
                "To see what files are staged, run: git status"
            )

        # Existing branches
        local_branches = self.local_branches_dict().keys()
        remote_branches = self.remote_branches_dict()
        all_branches = local_branches | set(remote_branches.keys())

        # Create new branch, if specified
        if new_branch:
            if branch_name in all_branches:
                raise PayuBranchError(
                    f"A branch named {branch_name} already exists. "
                    "To checkout this branch, remove the new branch flag '-b' "
                    "from the checkout command."
                )

            if start_point is not None:
                if (start_point not in local_branches and
                        start_point in remote_branches):
                    # Use hash for remote start point
                    start_point = remote_branches[start_point].commit
                branch = self.repo.create_head(branch_name, commit=start_point)
            else:
                branch = self.repo.create_head(branch_name)
            branch.checkout()

            print(f"Created and checked out new branch: {branch_name}")
            return

        # Checkout branch
        if branch_name not in all_branches:
            raise PayuBranchError(
                f"There is no existing branch called {branch_name}. "
                "To create this branch, add the new branch flag '-b' "
                "to the checkout command."
            )

        self.repo.git.checkout(branch_name)
        print(f"Checked out branch: {branch_name}")


def git_clone(repository: str,
              directory: Union[str, Path],
              branch: Optional[str] = None) -> GitRepository:
    """Clone repository to directory"""
    # Clone the repository
    if branch is not None:
        repo = git.Repo.clone_from(repository,
                                   to_path=directory,
                                   branch=branch)
    else:
        repo = git.Repo.clone_from(repository, to_path=directory)

    print(f"Cloned repository from {repository} to directory: {directory}")

    return GitRepository(repo_path=directory, repo=repo)



# coding: utf-8
"""payu.models.oasis
   =================

   Driver interface to the oasis coupler.

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""

# Standard Library
import os
import shutil

# Extensions
import f90nml

# Local
from payu.fsops import mkdir_p, make_symlink
from payu.models.model import Model
from payu.namcouple import Namcouple


class Oasis(Model):

    def __init__(self, expt, name, config):
        super(Oasis, self).__init__(expt, name, config)

        self.model_type = 'oasis'
        self.copy_restarts = True
        self.copy_inputs = False

        self.config_files = ['namcouple']

    def setup(self):
        super(Oasis, self).setup()

        # Copy OASIS data to the other submodels

        # TODO: Parse namcouple to determine filelist
        # TODO: Let users map files to models
        input_files = [f for f in os.listdir(self.work_path)
                       if f not in self.config_files]

        for model in self.expt.models:

            # Skip the oasis self-reference
            if model == self:
                continue

            # Skip models without a work_path (like access)
            if not hasattr(model, 'work_path'):
                continue

            mkdir_p(model.work_path)
            for f_name in (self.config_files + input_files):
                f_path = os.path.join(self.work_path, f_name)
                f_sympath = os.path.join(model.work_path, f_name)
                make_symlink(f_path, f_sympath)

        if self.expt.runtime:
            # TODO: Implement runtime patch to namcouple
            pass

    def set_timestep(self, t_step):

        namcpl_path = os.path.join(self.work_path, 'namcouple')
        namcpl = Namcouple(namcpl_path, 'access')
        namcpl.set_ice_ocean_coupling_timestep(str(t_step))
        namcpl.write()

        for model in self.expt.models:

            if model.model_type in ('cice', 'cice5'):

                # Set namcouple timesteps

                ice_ts = model.config.get('timestep')
                if ice_ts:
                    model.set_oasis_timestep(ice_ts)

                # Set ACCESS coupler timesteps

                input_ice_path = os.path.join(model.work_path, 'input_ice.nml')
                input_ice = f90nml.read(input_ice_path)

                input_ice['coupling_nml']['dt_cpl_io'] = t_step

                input_ice.write(input_ice_path, force=True)

            elif model.model_type == 'matm':

                input_atm_path = os.path.join(model.work_path, 'input_atm.nml')
                input_atm = f90nml.read(input_atm_path)

                input_atm['coupling']['dt_atm'] = t_step

                input_atm.write(input_atm_path, force=True)

            elif model.model_type == 'mom':

                input_nml_path = os.path.join(model.work_path, 'input.nml')
                input_nml = f90nml.read(input_nml_path)

                input_nml['auscom_ice_nml']['dt_cpl'] = t_step
                input_nml['ocean_solo_nml']['dt_cpld'] = t_step

                input_nml.write(input_nml_path, force=True)

    def archive(self):
        super(Oasis, self).archive()

        # TODO: Determine the exchange files
        restart_files = ['a2i.nc', 'i2a.nc', 'i2o.nc', 'o2i.nc']

        mkdir_p(self.restart_path)
        for f in restart_files:
            f_src = os.path.join(self.work_path, f)
            f_dst = os.path.join(self.restart_path, f)

            if os.path.exists(f_src):
                shutil.move(f_src, f_dst)

    def collate(self):
        pass



"""
The payu interface for the CICE model
-------------------------------------------------------------------------------
Contact: Marshall Ward <marshall.ward@anu.edu.au>
-------------------------------------------------------------------------------
Distributed as part of Payu, Copyright 2011 Marshall Ward
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""

from __future__ import print_function

import os

from payu.models.cice import Cice
from payu.fsops import mkdir_p


class Cice5(Cice):

    def __init__(self, expt, name, config):
        super(Cice5, self).__init__(expt, name, config)

        self.model_type = 'cice5'
        self.default_exec = 'cice'

        # Default repo details
        self.repo_url = 'https://github.com/OceansAus/cice5.git'
        self.repo_tag = 'master'

        self.config_files = [
            'cice_in.nml',
            'input_ice.nml',
            'input_ice_gfdl.nml',
            'input_ice_monin.nml'
        ]

        self.ice_nml_fname = 'cice_in.nml'

        self.set_timestep = self.set_local_timestep

        self.copy_restarts = True
        self.copy_inputs = True

        # Empty list means no log files will be compressed
        self.logs_to_compress = []

    def set_local_timestep(self, t_step):
        dt = self.ice_in['setup_nml']['dt']
        npt = self.ice_in['setup_nml']['npt']

        self.ice_in['setup_nml']['dt'] = t_step
        self.ice_in['setup_nml']['npt'] = (int(dt) * int(npt)) // int(t_step)

        ice_in_path = os.path.join(self.work_path, self.ice_nml_fname)
        self.ice_in.write(ice_in_path, force=True)

    def setup(self):
        # Force creation of a dump (restart) file at end of run
        self.ice_in['setup_nml']['dump_last'] = True

        super(Cice5, self).setup()

        # Make log dir
        mkdir_p(os.path.join(self.work_path, 'log'))

    def get_prior_restart_files(self):
        if self.prior_restart_path is not None:
            return sorted(os.listdir(self.prior_restart_path))
        else:
            return []

    def set_access_timestep(self, t_step):
        # TODO: Figure out some way to move this to the ACCESS driver
        # Re-read ice timestep and move this over there
        self.set_local_timestep(t_step)

    def _calc_runtime(self):
        """
        Overrides the cice driver method, as CICE5 can store the timing information in restart files does not use
        the timing information in the cice_in.nml namelist.
        """
        pass

    def _make_restart_ptr(self):
        """
        Generate restart pointer which points to the latest iced.YYYYMMDD
        restart file.
        """
        iced_restart_file = None
        iced_restart_files = [f for f in self.get_prior_restart_files()
                              if f.startswith('iced.')]

        if len(iced_restart_files) > 0:
            iced_restart_file = sorted(iced_restart_files)[-1]

        if iced_restart_file is None:
            raise FileNotFoundError(
                f'No iced restart file found in {self.prior_restart_path}')

        res_ptr_path = os.path.join(self.work_init_path,
                                    'ice.restart_file')
        if os.path.islink(res_ptr_path):
            # If we've linked in a previous pointer it should be deleted
            os.remove(res_ptr_path)
        with open(res_ptr_path, 'w') as res_ptr:
            res_dir = self.get_ptr_restart_dir()
            res_ptr.write(os.path.join(res_dir, iced_restart_file))



import shutil
import subprocess

import git
import pytest

from payu.git_utils import get_git_repository, GitRepository
from payu.git_utils import PayuBranchError, PayuGitWarning

from test.common import tmpdir


@pytest.fixture(autouse=True)
def setup_and_teardown():
    # Create tmp directory
    try:
        tmpdir.mkdir()
    except Exception as e:
        print(e)

    yield

    # Remove tmp directory
    try:
        shutil.rmtree(tmpdir)
    except Exception as e:
        print(e)


def create_new_repo(repo_path):
    """Helper function to initialise a repo and create first commit"""
    repo = git.Repo.init(repo_path)
    init_file = repo_path / "init.txt"
    add_file_and_commit(repo, init_file)
    return repo


def add_file_and_commit(repo, file_path, commit_no=0):
    """Helper function to add a commit to repo"""
    file_path.touch()
    repo.index.add([file_path])
    repo.index.commit(f"Add commit {commit_no}")
    return repo


def test_get_git_repo_invalid_repo_initialise():
    invalid_repo_path = tmpdir / "invalidRepo"
    invalid_repo_path.mkdir()
    repo = get_git_repository(invalid_repo_path, initialise=True)
    assert not repo.bare


def test_get_git_repo_invalid_repo_catch_error():
    invalid_path = tmpdir / "invalidRepo"
    invalid_path.mkdir()
    expected_warning_msg = "Path is not a valid git repository: "
    expected_warning_msg += str(invalid_path)
    with pytest.warns(PayuGitWarning, match=expected_warning_msg):
        repo = get_git_repository(invalid_path, catch_error=True)
        assert repo is None


def test_get_git_user_info_no_config_set():
    # Testing this is tricky as don't want to remove any global configs for
    # name or email. Instead using assumption that key 'testKey-54321' is not
    # defined in the 'user' namespace.
    repo_path = tmpdir / "test_repo"
    create_new_repo(repo_path)
    repo = GitRepository(repo_path)
    value = repo.get_user_info('testKey-54321')
    assert value is None


def test_get_git_user_info_config_set():
    repo_path = tmpdir / "test_repo"
    create_new_repo(repo_path)
    try:
        # Set config that is local to temporary test repository only
        subprocess.run('git config user.name "TestUserName"',
                       check=True,
                       shell=True,
                       cwd=repo_path)
        print("User name set successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error setting user name: {e}")

    repo = GitRepository(repo_path)
    value = repo.get_user_info('name')

    assert value == 'TestUserName'


@pytest.mark.parametrize("ref", ["branch", "hash", None])
def test_git_checkout_new_branch_from_remote_ref(ref):
    # Setup
    remote_repo_path = tmpdir / 'remoteRepo'
    remote_repo = create_new_repo(remote_repo_path)
    main_branch = remote_repo.active_branch
    main_branch_hash = main_branch.object.hexsha

    # Create branch_1
    branch_1 = remote_repo.create_head("branch-1")
    remote_repo.git.checkout(branch_1)
    add_file_and_commit(remote_repo, (remote_repo_path / 'file'), commit_no=1)
    branch_1_hash = branch_1.object.hexsha

    assert main_branch_hash != branch_1_hash

    # Re-checkout main branch
    remote_repo.git.checkout(main_branch)

    # Clone repo
    cloned_repo_path = tmpdir / 'cloned_repo'
    cloned_repo = remote_repo.clone(cloned_repo_path)

    if ref == "hash":
        start_point = branch_1_hash
        expected_hash = branch_1_hash
    elif ref == "branch":
        start_point = "branch-1"
        expected_hash = branch_1_hash
    else:
        start_point = None
        expected_hash = main_branch_hash

    # Test startpoint being remote branch/hash/None
    repo = GitRepository(cloned_repo_path)
    repo.checkout_branch('branch-2',
                         new_branch=True,
                         start_point=start_point)

    current_branch = cloned_repo.active_branch
    current_hash = current_branch.object.hexsha
    assert str(current_branch) == 'branch-2'
    assert current_hash == expected_hash


def test_git_checkout_new_branch_existing():
    # Setup
    repo_path = tmpdir / 'remoteRepo'
    repo = create_new_repo(repo_path)
    existing_branch = repo.active_branch

    # Test checkout branch with existing branch
    repo = GitRepository(repo_path)
    with pytest.raises(PayuBranchError):
        repo.checkout_branch(str(existing_branch),
                             new_branch=True)


def test_git_checkout_non_existent_branch():
    # Setup
    repo_path = tmpdir / 'remoteRepo'
    create_new_repo(repo_path)

    # Test checkout branch with non-existent branch
    repo = GitRepository(repo_path)
    with pytest.raises(PayuBranchError):
        repo.checkout_branch("Gibberish")


def test_git_checkout_staged_changes():
    # Setup
    repo_path = tmpdir / 'remoteRepo'
    create_new_repo(repo_path)

    repo = GitRepository(repo_path)
    file_path = repo_path / 'newTestFile.txt'
    file_path.touch()

    # Test checkout branch works with untracked files
    repo.checkout_branch(new_branch=True, branch_name="NewBranch")

    # Test checkout raises error with staged changes
    repo.repo.index.add([file_path])
    with pytest.raises(PayuBranchError):
        repo.checkout_branch(new_branch=True, branch_name="NewBranch2")


def test_git_checkout_existing_branch():
    # Setup
    remote_repo_path = tmpdir / 'remoteRepo'
    remote_repo = create_new_repo(remote_repo_path)
    main_branch = remote_repo.active_branch

    # Create branch_1
    branch_1 = remote_repo.create_head("branch-1")
    remote_repo.git.checkout(branch_1)
    add_file_and_commit(remote_repo, (remote_repo_path / 'file'), commit_no=1)
    branch_1_hash = branch_1.object.hexsha

    # Re-checkout main branch
    remote_repo.git.checkout(main_branch)

    # Clone repo
    cloned_repo_path = tmpdir / 'cloned_repo'
    cloned_repo = remote_repo.clone(cloned_repo_path)

    # Test checkout existing remote branch
    repo = GitRepository(cloned_repo_path)
    repo.checkout_branch('branch-1')

    current_branch = cloned_repo.active_branch
    current_hash = current_branch.object.hexsha
    assert str(current_branch) == 'branch-1'
    assert current_hash == branch_1_hash


def test_git_checkout_missing_origin_repo():
    """In Issue #405, there was a bug where origin remote path was deleted,
    the payu checkout would raise an error when fetching remote repository"""
    # Create remote repository
    remote_repo_path = tmpdir / 'remote_repo'
    remote_repo = create_new_repo(remote_repo_path)

    # Clone repository
    cloned_repo_path = tmpdir / 'cloned_repo'
    remote_repo.clone(cloned_repo_path)

    # Remove remote directory - this will raise errors when remote is fetched
    shutil.rmtree(remote_repo_path)

    repo = GitRepository(cloned_repo_path)

    # Check remote branches runs with a warning
    with pytest.warns(PayuGitWarning):
        repo.checkout_branch(branch_name="branch-1", new_branch=True)

        # No remote branches found
        assert repo.remote_branches_dict() == {}



"""Functions to support Slurm scheduling.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

# Standard library
import os
import re
import sys
import shlex
import subprocess

from payu.fsops import check_exe_path
from payu.schedulers.scheduler import Scheduler


class Slurm(Scheduler):
    # TODO: __init__

    def submit(self, pbs_script, pbs_config, pbs_vars=None, python_exe=None):
        """Prepare a correct PBS command string"""

        if python_exe is None:
            python_exe = sys.executable

        if pbs_vars is None:
            pbs_vars = {}

        payu_path = pbs_vars.get('PAYU_PATH', os.path.dirname(sys.argv[0]))
        pbs_script = check_exe_path(payu_path, pbs_script)

        pbs_flags = []
        pbs_flags.append('--time={}'.format(pbs_config.get('walltime')))
        pbs_flags.append('--ntasks={}'.format(pbs_config.get('ncpus')))

        # Flags which need to be addressed
        pbs_flags.append('--qos=debug')
        pbs_flags.append('--cluster=c4')

        # Construct job submission command
        cmd = 'sbatch {flags} --wrap="{python} {script}"'.format(
            flags=' '.join(pbs_flags),
            python=python_exe,
            script=pbs_script
        )

        return cmd



"""Generic model interface, primarily to be inherited by other models.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
:license: Apache License, Version 2.0, see LICENSE for details
"""

import os
import shutil
import shlex
import sys
import subprocess as sp

from payu import envmod
from payu.fsops import mkdir_p, required_libs


class Model(object):
    """Abstract model class."""

    def __init__(self, expt, model_name, model_config):
        """Create the model interface."""
        # Inherit experiment configuration
        self.expt = expt
        self.name = model_name
        self.config = model_config
        self.top_level_model = False

        # Model details
        self.model_type = None
        self.default_exec = None
        self.input_basepath = None
        self.modules = []
        self.config_files = []
        self.optional_config_files = []

        # Path names
        self.work_input_path = None
        self.work_restart_path = None
        self.work_init_path = None
        # A string to add before the exe name, useful for e.g. gdb, gprof,
        # valgrind
        self.exec_prefix = None
        self.exec_path = None
        self.exec_name = None
        self.codebase_path = None
        self.work_path_local = None
        self.work_input_path_local = None
        self.work_restart_path_local = None
        self.work_init_path_local = None
        self.exec_path_local = None

        self.build_exec_path = None
        self.build_path = None

        self.required_libs = None

        # Control flags
        self.copy_restarts = False
        self.copy_inputs = False

        # Codebase details
        self.repo_url = None
        self.repo_tag = None
        self.build_command = None

    def set_model_pathnames(self):
        """Define the paths associated with this model."""
        self.control_path = self.expt.control_path
        self.input_basepath = self.expt.lab.input_basepath
        self.work_path = self.expt.work_path
        self.codebase_path = self.expt.lab.codebase_path

        if len(self.expt.models) > 1:

            self.control_path = os.path.join(self.control_path, self.name)
            self.work_path = os.path.join(self.work_path, self.name)
            self.codebase_path = os.path.join(self.codebase_path, self.name)

        # NOTE: Individual models may override the work subdirectories
        self.work_input_path = self.work_path
        self.work_restart_path = self.work_path
        self.work_output_path = self.work_path
        self.work_init_path = self.work_path

    def set_local_pathnames(self):

        # This is the path relative to the control directory, required for
        # manifests and must be called after set_model_pathnames to ensure it
        # captures changes made in model subclasses which override
        # set_model_pathnames

        # XXX: If path is relative to control_path, why not use it?
        self.work_path_local = os.path.normpath(
            os.path.join(
                'work',
                os.path.relpath(self.work_path, self.expt.work_path)
            )
        )
        self.work_input_path_local = os.path.normpath(
            os.path.join(
                'work',
                os.path.relpath(self.work_input_path, self.expt.work_path)
            )
        )
        self.work_restart_path_local = os.path.normpath(
            os.path.join(
                'work',
                os.path.relpath(self.work_restart_path, self.expt.work_path)
            )
        )
        self.work_init_path_local = os.path.normpath(
            os.path.join(
                'work',
                os.path.relpath(self.work_init_path, self.expt.work_path)
            )
        )

    def set_input_paths(self):
        if len(self.expt.models) == 1:
            input_dirs = self.expt.config.get('input')
        else:
            input_dirs = self.config.get('input')

        if input_dirs is None:
            input_dirs = []
        elif isinstance(input_dirs, str):
            input_dirs = [input_dirs]

        self.input_paths = []
        for input_dir in input_dirs:

            # First test if path exists
            if os.path.exists(input_dir):
                # Resolve symlinks and relative paths
                input_dir = os.path.realpath(input_dir)

                self.input_paths.append(input_dir)
            else:
                # Test for path relative to /${lab_path}/input/${model_name}
                assert self.input_basepath
                rel_path = os.path.join(self.input_basepath, input_dir)
                if os.path.exists(rel_path):
                    # Resolve symlinks
                    rel_path = os.path.realpath(rel_path)

                    self.input_paths.append(rel_path)
                else:
                    sys.exit('payu: error: Input directory {0} not found; '
                             'aborting.'.format(rel_path))

    def set_model_output_paths(self):

        self.output_path = self.expt.output_path
        self.restart_path = self.expt.restart_path

        self.prior_output_path = self.expt.prior_output_path
        self.prior_restart_path = self.expt.prior_restart_path

        if len(self.expt.models) > 1:

            # If '-d' option specified for collate don't want to change the
            # output path, but respect the absolute path specified
            if os.environ.get('PAYU_DIR_PATH') is None:
                self.output_path = os.path.join(self.output_path, self.name)

            self.restart_path = os.path.join(self.restart_path, self.name)

            if self.prior_output_path:
                self.prior_output_path = os.path.join(self.prior_output_path,
                                                      self.name)

            if self.prior_restart_path:
                self.prior_restart_path = os.path.join(self.prior_restart_path,
                                                       self.name)

    def get_prior_restart_files(self):

        try:
            respath = self.prior_restart_path
            return [f for f in os.listdir(respath)
                    if os.path.isfile(os.path.join(respath, f))]
        except Exception as e:
            print("No prior restart files found: {error}".format(error=str(e)))
            return []

    def expand_executable_path(self, exec):
        """Given an executable, return the expanded executable path"""
        # Check if exe is already an absolute path
        if os.path.isabs(exec):
            return exec

        # Check if path set by loading user modules has been defined
        module_added_paths = self.expt.user_modules_paths
        if module_added_paths is None:
            print("payu: warning: Skipping searching for model executable " +
                  "in $PATH set by user modules")
            module_added_paths = []

        # Search for exe inside paths added to $PATH by user-defined modules
        exec_paths = []
        for path in module_added_paths:
            exec_path = os.path.join(path, exec)
            if os.path.exists(exec_path) and os.access(exec_path, os.X_OK):
                exec_paths.append(exec_path)

        if len(exec_paths) > 1:
            raise ValueError(
                f"Executable {exec} found in multiple $PATH paths added by " +
                f"user-defined modules in `config.yaml`. Paths: {exec_paths}")
        elif len(exec_paths) == 1:
            return exec_paths[0]

        # Else prepend the lab bin path to exec
        return os.path.join(self.expt.lab.bin_path, exec)

    def setup_executable_paths(self):
        """Set model executable paths"""
        self.exec_prefix = self.config.get('exe_prefix', '')
        self.exec_name = self.config.get('exe', self.default_exec)
        self.exec_path = None
        if self.exec_name:
            self.exec_path = self.expand_executable_path(self.exec_name)

            # Make exec_name consistent for models with fully qualified path.
            # In all cases it will just be the name of the executable without a
            # path
            self.exec_name = os.path.basename(self.exec_path)

            # Local path in work directory
            self.exec_path_local = os.path.join(
                self.work_path_local,
                os.path.basename(self.exec_path)
            )

    def setup_configuration_files(self):
        """Copy configuration and optional configuration files from control
         path to work path"""
        for f_name in self.config_files:
            f_path = os.path.join(self.control_path, f_name)
            shutil.copy(f_path, self.work_path, follow_symlinks=True)

        for f_name in self.optional_config_files:
            f_path = os.path.join(self.control_path, f_name)
            try:
                shutil.copy(f_path, self.work_path, follow_symlinks=True)
            except FileNotFoundError:
                # Ignore missing files for optional config files
                pass

    def setup(self):

        print("Setting up {model}".format(model=self.name))
        # Create experiment directory structure
        mkdir_p(self.work_init_path)
        mkdir_p(self.work_input_path)
        mkdir_p(self.work_restart_path)
        mkdir_p(self.work_output_path)

        # Copy configuration files from control path
        self.setup_configuration_files()

        # Add restart files from prior run to restart manifest
        if self.prior_restart_path:
            restart_files = self.get_prior_restart_files()
            for f_name in restart_files:
                f_orig = os.path.join(self.prior_restart_path, f_name)
                f_link = os.path.join(self.work_init_path_local, f_name)
                self.expt.manifest.add_filepath(
                    'restart',
                    f_link,
                    f_orig,
                    self.copy_restarts
                )

        # Add input files to input manifest
        for input_path in self.input_paths:
            if os.path.isfile(input_path):
                # Build a mock walk iterator for a single file
                fwalk = iter([(
                    os.path.dirname(input_path),
                    [],
                    [os.path.basename(input_path)]
                )])
                # Overwrite the input_path as a directory
                input_path = os.path.dirname(input_path)
            else:
                fwalk = os.walk(input_path)

            for path, dirs, files in fwalk:
                workrelpath = os.path.relpath(path, input_path)
                subdir = os.path.normpath(
                    os.path.join(self.work_input_path_local,
                                 workrelpath)
                )

                if not os.path.exists(subdir):
                    os.mkdir(subdir)

                for f_name in files:
                    f_orig = os.path.join(path, f_name)
                    f_link = os.path.join(
                        self.work_input_path_local,
                        workrelpath,
                        f_name
                    )
                    # Do not use input file if already linked
                    # as a restart file
                    if not os.path.exists(f_link):
                        self.expt.manifest.add_filepath(
                            'input',
                            f_link,
                            f_orig,
                            self.copy_inputs
                        )

        # Make symlink to executable in work directory
        if self.exec_path:
            # Check whether executable path exists
            if not os.path.isfile(self.exec_path):
                raise FileNotFoundError(
                    f'Executable for {self.name} model '
                    f'not found on path: {self.exec_path}')

            # Check whether executable has executable permission
            if not os.access(self.exec_path, os.X_OK):
                raise PermissionError(
                    f'Executable for {self.name} model '
                    f'is not executable: {self.exec_path}')

            # Add to exe manifest (this is always done so any change in exe
            # path will be picked up)
            self.expt.manifest.add_filepath(
                'exe',
                self.exec_path_local,
                self.exec_path
            )

            # Populate information about required dynamically loaded libraries
            self.required_libs = required_libs(self.exec_path)

        timestep = self.config.get('timestep')
        if timestep:
            self.set_timestep(timestep)

    def set_timestep(self, timestep):
        """Set the model timestep."""
        raise NotImplementedError

    def archive(self):
        """Store model output to laboratory archive."""

        # Traverse the model directory deleting symlinks, zero length files
        # and empty directories
        for path, dirs, files in os.walk(self.work_path, topdown=False):
            for f_name in files:
                f_path = os.path.join(path, f_name)
                if os.path.islink(f_path) or os.path.getsize(f_path) == 0:
                    os.remove(f_path)
            if len(os.listdir(path)) == 0:
                os.rmdir(path)

    def collate(self):
        """Collate any tiled output into a single file."""
        raise NotImplementedError

    def build_model(self):
        self.setup_executable_paths()

        if not self.repo_url:
            return

        # Check to see if executable already exists.
        if self.exec_path and os.path.isfile(self.exec_path):
            print('payu: warning: {0} will be overwritten.'
                  ''.format(self.exec_path))

        # First step is always to go to the codebase.
        curdir = os.getcwd()

        # Do the build. First check whether there is a build command in the
        # config. If not check for the model default, otherwise just run make.

        try:
            build_path = self.config['build']['path_to_build_command']
        except KeyError:
            if self.build_path:
                build_path = self.build_path
            else:
                build_path = './'

        os.chdir(os.path.join(self.codebase_path, build_path))

        try:
            cmd = self.config['build']['command']
        except KeyError:
            if self.build_command:
                cmd = self.build_command
            else:
                cmd = 'make'

        print('Running command {0}'.format(cmd))
        sp.check_call(shlex.split(cmd))

        try:
            build_exec_path = os.path.join(self.codebase_path,
                                           self.config['build']['exec_path'])
        except KeyError:
            if self.build_exec_path:
                build_exec_path = self.build_exec_path
            else:
                build_exec_path = self.codebase_path

        # Copy new executable to bin dir
        if self.exec_path:
            # Create the bin path if it doesn't exist
            mkdir_p(self.expt.lab.bin_path)

            build_exec_path = os.path.join(build_exec_path, self.exec_name)
            shutil.copy(build_exec_path, self.exec_path)

        os.chdir(curdir)

    def get_codebase(self):

        if not self.repo_url:
            return

        # Disable the user's .gitconfig file
        os.environ['GIT_CONFIG_NOGLOBAL'] = 'yes'

        build_config = self.config.get('build', {})
        self.repo_url = build_config.get('repository', self.repo_url)
        self.repo_tag = build_config.get('tag', self.repo_tag)

        git_path = os.path.join(self.codebase_path, '.git')
        if not os.path.exists(git_path):
            cmd = 'git clone {0} {1}'.format(self.repo_url, self.codebase_path)
            sp.check_call(shlex.split(cmd))

        curdir = os.getcwd()
        os.chdir(self.codebase_path)
        sp.check_call(shlex.split('git checkout {0}'.format(self.repo_tag)))
        sp.check_call(shlex.split('git pull'))
        os.chdir(curdir)

    def profile(self):
        # TODO: Replace with call to "profile" drivers

        if self.expt.config.get('hpctoolkit', False) and self.exec_name:

            envmod.module('load', 'hpctoolkit')

            # Create the code structure file
            hpcstruct_fname = '{0}.hpcstruct'.format(self.exec_name)
            hpcstruct_path = os.path.join(self.expt.lab.bin_path,
                                          hpcstruct_fname)

            # TODO: Validate struct file
            if not os.path.isfile(hpcstruct_path):
                cmd = 'hpcstruct -o {0} {1}'.format(hpcstruct_path,
                                                    self.exec_path)
                sp.check_call(shlex.split(cmd))

            # Parse the profile output
            hpctk_header = 'hpctoolkit-{0}-measurements'.format(self.exec_name)
            hpctk_measure_dir = [os.path.join(self.output_path, f)
                                 for f in os.listdir(self.output_path)
                                 if f.startswith(hpctk_header)][0]

            hpctk_db_dir = hpctk_measure_dir.replace('measurements',
                                                     'database')

            # TODO: This needs to be model-specifc
            src_path = os.path.join(self.codebase_path, 'src')

            cmd = 'hpcprof-mpi -S {0} -I {1} -o {2} {3}'.format(
                hpcstruct_path, src_path, hpctk_db_dir, hpctk_measure_dir)
            sp.check_call(shlex.split(cmd))

        if self.expt.config.get('scalasca', False):

            envmod.module('use', '/home/900/mpc900/my_modules')
            envmod.module('load', 'scalasca')

            scorep_path = [os.path.join(self.output_path, f)
                           for f in os.listdir(self.output_path)
                           if f.startswith('scorep')][0]
            cmd = 'scalasca -examine -s {0}'.format(scorep_path)
            sp.check_call(shlex.split(cmd))

        if self.expt.config.get('scorep', False):

            envmod.module('load', 'scorep')

            scorep_path = [os.path.join(self.output_path, f)
                           for f in os.listdir(self.output_path)
                           if f.startswith('scorep')][0]
            cube_path = [os.path.join(scorep_path, f)
                         for f in os.listdir(scorep_path)
                         if f.endswith('.cubex')][0]
            cmd = 'scorep-score {0}'.format(cube_path)
            sp.check_call(shlex.split(cmd))

    def get_restart_datetime(self, restart_path):
        """Given a restart path, parse the restart files and return a cftime
        datetime (currently used for date-based restart pruning)"""
        raise NotImplementedError



# coding: utf-8
"""
The payu interface for CESM model configurations using CMEPS and the CESM CMEPS driver
--------------------------------------------------------------------------------------
Contact: Marshall Ward <marshall.ward@anu.edu.au>
--------------------------------------------------------------------------------------
Distributed as part of Payu, Copyright 2011 Marshall Ward
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""
from __future__ import print_function

import os
import re
import shutil
from warnings import warn

from payu.fsops import mkdir_p, make_symlink
from payu.models.model import Model
from payu.models.fms import fms_collate
from payu.models.mom6 import mom6_add_parameter_files

NUOPC_CONFIG = "nuopc.runconfig"
NUOPC_RUNSEQ = "nuopc.runseq"

# Add as needed
component_info = {
    "mom": {
        "config_files": [
            "input.nml",
            "diag_table",
        ],
        "optional_config_files": [
            "field_table",
            "data_table",
        ],
    },
    "cice": {
        "config_files": ["ice_in"],
    },
    "ww3dev": {
        "config_files": ["wav_in"],
        "optional_config_files" : [
            "ww3_shel.nml",
            "ww3_points.list",
        ],
    },
    "datm": {
        "config_files": [
            "datm_in",
            "datm.streams.xml",
        ],
    },
    "docn": {
        "config_files": [
            "docn_in",
            "docn.streams.xml",
        ],
    },
    "drof": {
        "config_files": [
            "drof_in",
            "drof.streams.xml"
        ]
    },
}

class CesmCmeps(Model):

    def __init__(self, expt, name, config):
        super().__init__(expt, name, config)

        self.model_type = 'cesm-cmeps'

        # Driver config files should always be present
        self.config_files = [
            "drv_in",
            "fd.yaml",
            NUOPC_CONFIG,
            NUOPC_RUNSEQ
        ]

        self.realms = ["ocn", "ice", "wav", "atm", "rof", "cpl"]
        self.runconfig = None # nuopc.runconfig. Can't read this yet as paths haven't necessarily been set
        self.components = {} # To be read from nuopc.runconfig
        self.rpointers = [] # To be inferred from nuopc.runconfig

    def get_runconfig(self, path):
        self.runconfig = Runconfig(os.path.join(path, NUOPC_CONFIG))
        
    def get_components(self):
        """Get components from nuopc.runconfig"""

        self.components = {}
        self.rpointers = ["rpointer.cpl"]
        for realm in self.realms:
            if realm != "cpl":
                component = self.runconfig.get("ALLCOMP_attributes", f"{realm.upper()}_model")
                if not component.startswith("s"): # TODO: better check for stub component?
                    self.components[realm] = component
                    self.rpointers.append(f"rpointer.{realm}")

        for component in self.components.values():
            if component in component_info:
                self.config_files.extend(component_info[component]["config_files"])
                try:
                    self.optional_config_files.extend(
                        component_info[component]["optional_config_files"]
                        )
                except KeyError:
                    pass
            else:
                raise ValueError(f"There is no component info entry for {component}")

    def set_model_pathnames(self):

        super().set_model_pathnames()

        self.work_input_path = os.path.join(self.work_path, 'INPUT')
        
        # MOM restarts are dealt with via pointer files (see below). Use work_restart_path
        # for additional restarts (e.g. generic tracer flux restarts)
        self.get_runconfig(self.control_path)
        additional_restart_dir = self.runconfig.get(
            "ALLCOMP_attributes", "additional_restart_dir", "RESTART"
        )
        self.work_restart_path = os.path.join(self.work_path, additional_restart_dir)

    def setup(self):
        super().setup()

        # Read components from nuopc.runconfig
        self.get_components()

        if "mom" in self.components.values():
            # Read mom6 parameter files and add to configuration files
            mom6_add_parameter_files(self)

        # Copy configuration files from control path to work path
        self.setup_configuration_files()

        if self.prior_restart_path and not self.expt.repeat_run:
            start_type = 'continue'

            # Overwrite restart pointer symlinks with copies
            pointer_files = [
                os.path.join(
                    self.work_path, pointer,
                    ) for pointer in self.rpointers
            ]
            for f_dst in pointer_files:
                f_src = os.readlink(f_dst)
                os.remove(f_dst)
                shutil.copy(f_src, f_dst)
        else:
            start_type = 'startup'

        self.runconfig.set("ALLCOMP_attributes", "start_type", start_type)

        # run checks on nuopc.runfig
        self._setup_checks()

        # Ensure that restarts will be written at the end of each run
        stop_n = self.runconfig.get("CLOCK_attributes", "stop_n")
        stop_option = self.runconfig.get("CLOCK_attributes", "stop_option")
        self.runconfig.set("CLOCK_attributes", "restart_n", stop_n)
        self.runconfig.set("CLOCK_attributes", "restart_option", stop_option)
        
        mkdir_p(os.path.join(self.work_path, 'log'))
        mkdir_p(os.path.join(self.work_path, 'timing'))

        self.runconfig.write(os.path.join(self.work_path, NUOPC_CONFIG))

        # Horrible hack to make a link to the mod_def.ww3 input in the work
        # directory
        # The ww3 mod_def input needs to be in work_path and called mod_def.ww3
        if "ww3dev" in self.components.values():
            f_name = "mod_def.ww3"
            f_src = os.path.join(self.work_input_path, f_name)
            f_dst = os.path.join(self.work_path, f_name)

            if os.path.isfile(f_src):
                make_symlink(f_src, f_dst)
            else:
                # TODO: copied this from other models. Surely we want to exit here or something
                print('payu: error: Unable to find mod_def.ww3 file in input directory')

    def _setup_checks(self):
        # check pelayout fits within requested cpucount
        cpucount = int(self.expt.config.get('ncpus'))
        all_realms = self.realms
        for realm in all_realms:
            ntasks = int(self.runconfig.get("PELAYOUT_attributes", f"{realm}_ntasks"))
            nthreads = int(self.runconfig.get("PELAYOUT_attributes", f"{realm}_nthreads"))
            rootpe = int(self.runconfig.get("PELAYOUT_attributes", f"{realm}_rootpe"))
            pestride = int(self.runconfig.get("PELAYOUT_attributes", f"{realm}_pestride"))

            if nthreads < 1:
                raise ValueError(f"The number of {realm}_nthreads ({nthreads}) in "
                                 f"{NUOPC_CONFIG} must be at least 1.")
            if nthreads > 1:
                npes = nthreads*ntasks*pestride
                # this is taken from
                # https://github.com/ESCOMP/CMEPS/blob/5b7d76978e2fdc661ec2de4ba9834b985decadc6/cesm/driver/esm.F90#L1007
                # the correct calculation might be (ntasks-1)*pestride*nthreads + nthreads
            else:
                npes = (ntasks-1)*pestride + 1

            if (rootpe + npes) > cpucount:
                raise ValueError(
                    f"Insufficient cpus for the {realm} pelayout in {NUOPC_CONFIG}"
                )

            # check iolayout
            if realm == "cpl" or realm == "med":
                comp = "MED"  # med and cpl names are both used in runconfig
            else:
                comp = realm.upper()

            if comp in self.runconfig.get_component_list():
                io_section = f"{comp}_modelio"
                nc_type = self.runconfig.get(io_section, "pio_typename")
                ioroot = int(self.runconfig.get(io_section, "pio_root"))

                if ioroot >= npes:
                    raise ValueError(
                        f"{io_section} pio_root exceeds available PEs (max: {npes - 1}) "
                        f"in {NUOPC_CONFIG}."
                    )

                pio_async = self.runconfig.get(io_section, "pio_async_interface")
                if pio_async == ".true.":
                    warn(
                        "Payu does not do consistency checks for asynchronous pio, as "
                        f"set in {io_section} of {NUOPC_CONFIG}. Consider adding them"
                    )
                elif pio_async == ".false.":
                    match nc_type:
                        case "netcdf":
                            break
                        case "netcdf4p" | "pnetcdf":
                            niotasks = int(self.runconfig.get(io_section, "pio_numiotasks"))
                            iostride = int(self.runconfig.get(io_section, "pio_stride"))
                            if (niotasks <= 0):
                                warn(f"The pio_numiotasks for {io_section} in {NUOPC_CONFIG} is "
                                     "not set, using model default")
                            if (iostride <= 0):
                                warn(f"The pio_stride for {io_section} in {NUOPC_CONFIG} is "
                                     "not set, using model default")
                            if (all([
                                    niotasks > 0, 
                                    iostride > 0,
                                    (ioroot + (niotasks-1)*iostride) >= npes
                            ])):
                                raise ValueError(
                                    f"The iolayout for {io_section} in {NUOPC_CONFIG} is "
                                    "requesting out of range cpus"
                                )
                        case "netcdf4c":
                            raise ValueError(
                                f"netcdf4c in {io_section} of {NUOPC_CONFIG} is deprecated, "
                                "use netcdf4p"
                            )
                        case _:
                            raise ValueError(
                                f"The iotype for {io_section} in {NUOPC_CONFIG} is "
                                'invalid, valid options: "netcdf", "pnetcdf", "netcdf4p"'
                            )
        return True

    def archive(self):
        super().archive()

        mkdir_p(self.restart_path)

        # WW3 doesn't generate a rpointer file. Write one so that all components can be generally
        # handled in the same way.
        # Note, I don't actually know how WW3 knows what restart to use. The normal approach to
        # providing restarts to WW3 is to add a restart.ww3 file to the run directory. However, here
        # WW3 looks for a <case_name>.ww3.r.<date> file despite this not being specified in wav_in.
        # I guess the driver specifies this somewhere based on the cpl restart file name?
        if "ww3dev" in self.components.values():
            cpl_pointer = os.path.join(
                self.work_path, "rpointer.cpl",
                )
            with open(cpl_pointer, "r") as f:
                cpl_restart = f.readline().rstrip()
            wav_restart = cpl_restart.replace("cpl", "ww3").removesuffix(".nc")
            wav_pointer = os.path.join(
                self.work_path, "rpointer.wav",
                )
            with open(wav_pointer, 'w') as f:
                f.write(wav_restart)

        # Get names of restarts from restart pointer files
        pointer_files = [
                os.path.join(
                    self.work_path, pointer,
                    ) for pointer in self.rpointers
            ]
        restart_files = []
        for pointer in pointer_files:
            try:
                with open(pointer,"r") as f:
                    for restart in f.readlines():
                        restart_files.append(
                            os.path.join(
                                self.work_path, restart.rstrip(),
                                )
                            )
            except FileNotFoundError:
                # TODO: copied this from other models. Surely we want to exit here or something
                print(f"payu: error: Model has not produced pointer file {pointer}")
        
        for f_src in restart_files + pointer_files:
            name = os.path.basename(f_src)
            f_dst = os.path.join(self.restart_path, name)
            shutil.move(f_src, f_dst)

        # Archive any additional restarts in work_restart_path
        if os.path.exists(self.work_restart_path):
            for name in os.listdir(self.work_restart_path):
                f_src = os.path.join(self.work_restart_path, name)
                f_dst = os.path.join(self.restart_path, name)
                shutil.move(f_src, f_dst)
            os.rmdir(self.work_restart_path)

    def collate(self):
        
        # .setup is not run when collate is called so need to get components
        self.get_runconfig(self.output_path)
        self.get_components()
        
        if "mom" in self.components.values():
            fms_collate(self)
        else:
            super().collate()


class AccessOm3(CesmCmeps):

    def get_components(self):
        super().get_components()

        assert self.components["atm"] == "datm", (
            "Access-OM3 comprises a data atmosphere model, but the atmospheric model in nuopc.runconfig is set "
            f"to {self.components['atm']}."
        )
        assert self.components["rof"] == "drof", (
            "Access-OM3 comprises a data runoff model, but the runoff model in nuopc.runconfig is set "
            f"to {self.components['rof']}."
        )


class Runconfig:
    """ Simple class for parsing and editing nuopc.runconfig """

    def __init__(self, file):
        self.file = file
        with open(self.file, 'r+') as f:
            self.contents = f.read()

    def _get_variable_span(self, section, variable):
        # This matches everything between a section table heading
        # '{section}::' and a following '::'
        m = re.search(
            r"(?<={}\:\:)(.*?)(?=\:\:)".format(section),
            self.contents,
            re.DOTALL,
        )
        if m is not None:
            section_start = m.start(1)
            section_end = m.end(1)
            # This matches everything between the equals sign in
            # a variable assignment and a # or newline character,
            # excluding whitespace
            m = re.search(
                r"{}\s*=\s*(\S+)\s*(?=(#|\n))".format(variable),
                self.contents[section_start:section_end],
            )
            if m is not None:
                return section_start + m.start(1), section_start + m.end(1)

    def get(self, section, variable, value=None):
        """
        Get the value of a variable, returning `value` if the variable does not exist
        """
        span = self._get_variable_span(section, variable)
        if span:
            return self.contents[span[0]:span[1]]
        else:
            return value

    def get_component_list(self):
        """
        Get the `component_list`
        """
        m = re.search(
            r'component_list:\s*(.*)',
            self.contents
        )

        if m is not None:
            components_str = m.group(1).strip()
            return components_str.split()
        else:
            return None

    def set(self, section, variable, new_value):
        """
        Overwrite the value of any existing variable
        """
        span = self._get_variable_span(section, variable)
        if span:
            self.contents = self.contents[:span[0]] + new_value + self.contents[span[1]:]
        else:
            raise NotImplementedError(
                "Cannot set value of variable that does not already exist in nuopc.runconfig"
            )

    def write(self, file=None):
        """
        Write the nuopc.runconfig to a file. If no file is provided, overwrite existing
        file
        """
        file = file or self.file
        with open(file, 'w') as f:
            f.write(self.contents)




# coding: utf-8
"""envmodules
   ==========

   A modular port of the Environment Modules Python ``init`` script
"""

import os
import shlex
import subprocess

# Python 2.6 subprocess.check_output support
if not hasattr(subprocess, 'check_output'):
    from backports import check_output
    subprocess.check_output = check_output

DEFAULT_BASEPATH = '/opt/Modules'
DEFAULT_VERSION = 'v4.3.0'

MODULE_NOT_FOUND_HELP = """ To fix module not being found:
- Check module name and version in config.yaml (listed under `modules: load:`)
- If module is found in a module directory, ensure this path is listed in
config.yaml under `modules: use:`, or run `module use` command prior to running
payu commands.
"""

MULTIPLE_MODULES_HELP = """ To fix having multiple modules available:
- Add version to the module in config.yaml (under `modules: load:`)
- Modify module directories in config.yaml (under `modules: use:`)
- Or modify module directories in user environment by using module use/unuse
commands, e.g.:
    $ module use dir # Add dir to $MODULEPATH
    $ module unuse dir # Remove dir from $MODULEPATH
"""


def setup(basepath=DEFAULT_BASEPATH):
    """Set the environment modules used by the Environment Module system."""

    module_version = os.environ.get('MODULE_VERSION', DEFAULT_VERSION)

    moduleshome = os.environ.get('MODULESHOME', None)

    if moduleshome is None:
        moduleshome = os.path.join(basepath, module_version)

    # Abort if MODULESHOME does not exist
    if not os.path.isdir(moduleshome):
        print('payu: warning: MODULESHOME does not exist; disabling '
              'environment modules.')
        try:
            del(os.environ['MODULESHOME'])
        except KeyError:
            pass
        return
    else:
        print('payu: Found modules in {}'.format(moduleshome))

    os.environ['MODULE_VERSION'] = module_version
    os.environ['MODULE_VERSION_STACK'] = module_version
    os.environ['MODULESHOME'] = moduleshome

    if 'MODULEPATH' not in os.environ:
        module_initpath = os.path.join(moduleshome, 'init', '.modulespath')
        with open(module_initpath) as initpaths:
            modpaths = [
                line.partition('#')[0].strip()
                for line in initpaths.readlines() if not line.startswith('#')
            ]

        os.environ['MODULEPATH'] = ':'.join(modpaths)

    os.environ['LOADEDMODULES'] = os.environ.get('LOADEDMODULES', '')

    # Environment modules with certain characters will cause corruption
    # when MPI jobs get launched on other nodes (possibly a PBS issue).
    #
    # Bash processes obscure the issue at NCI, since it occurs in an
    # environment module function, and bash moves those to the end of
    # the environment variable list.
    #
    # NCI's mpirun wrapper is a bash script, and therefore "fixes" by doing
    # the shuffle and limiting the damage to other bash functions, but some
    # wrappers (e.g. OpenMPI 2.1.x) may not be present.  So we manually patch
    # the problematic variable here.  But a more general solution would be nice
    # someday.

    if 'BASH_FUNC_module()' in os.environ:
        bash_func_module = os.environ['BASH_FUNC_module()']
        os.environ['BASH_FUNC_module()'] = bash_func_module.replace('\n', ';')


def module(command, *args):
    """Run the modulecmd tool and use its Python-formatted output to set the
    environment variables."""

    if 'MODULESHOME' not in os.environ:
        print('payu: warning: No Environment Modules found; skipping {0} call.'
              ''.format(command))
        return

    modulecmd = ('{0}/bin/modulecmd'.format(os.environ['MODULESHOME']))

    cmd = '{0} python {1} {2}'.format(modulecmd, command, ' '.join(args))

    envs, _ = subprocess.Popen(shlex.split(cmd),
                               stdout=subprocess.PIPE).communicate()
    exec(envs)


def lib_update(required_libs, lib_name):
    # Local import to avoid reversion interference
    # TODO: Bad design, fixme!
    # NOTE: We may be able to move this now that reversion is going away
    from payu import fsops

    for lib_filename, lib_path in required_libs.items():
        if lib_filename.startswith(lib_name) and lib_path.startswith('/apps/'): 
            # Load nci's /apps/ version of module if required 
            # pylint: disable=unbalanced-tuple-unpacking
            mod_name, mod_version = fsops.splitpath(lib_path)[2:4]

            module('unload', mod_name)
            module('load', os.path.join(mod_name, mod_version))
            return '{0}/{1}'.format(mod_name, mod_version)

    # If there are no libraries, return an empty string
    return ''


def setup_user_modules(user_modules, user_modulepaths):
    """Run module use + load commands for user-defined modules. Return
    tuple containing a set of loaded modules and paths added to
    LOADEDMODULES and PATH environment variable, as result of
    loading user-modules"""

    if 'MODULESHOME' not in os.environ:
        print(
            'payu: warning: No Environment Modules found; ' +
            'skipping running module use/load commands for any module ' +
            'directories/modulefiles defined in config.yaml')
        return (None, None)

    # Add user-defined directories to MODULEPATH
    for modulepath in user_modulepaths:
        if not os.path.isdir(modulepath):
            raise ValueError(
                f"Module directory is not found: {modulepath}" +
                "\n Check paths listed under `modules: use:` in config.yaml")

        module('use', modulepath)

    # First un-load all user modules, if they are loaded, so can store
    # LOADEDMODULES and PATH to compare to later
    for modulefile in user_modules:
        if run_module_cmd("is-loaded", modulefile).returncode == 0:
            module('unload', modulefile)
    previous_loaded_modules = os.environ.get('LOADEDMODULES', '')
    previous_path = os.environ.get('PATH', '')

    for modulefile in user_modules:
        # Check modulefile exists and is unique or has an exact match
        check_modulefile(modulefile)

        # Load module
        module('load', modulefile)

    # Create a set of paths and modules loaded by user modules
    loaded_modules = os.environ.get('LOADEDMODULES', '')
    path = os.environ.get('PATH', '')
    loaded_modules = set(loaded_modules.split(':')).difference(
        previous_loaded_modules.split(':'))
    paths = set(path.split(':')).difference(set(previous_path.split(':')))

    return (loaded_modules, paths)


def check_modulefile(modulefile: str) -> None:
    """Given a modulefile, check if modulefile exists, and there is
    a unique modulefile available - e.g. if it's version is specified"""
    output = run_module_cmd("avail --terse", modulefile).stderr

    # Extract out the modulefiles available - strip out lines like:
    # /apps/Modules/modulefiles:
    modules_avail = [line for line in output.strip().splitlines()
                     if not (line.startswith('/') and line.endswith(':'))]

    # Remove () from end of modulefiles if they exist, e.g. (default)
    modules_avail = [mod.rsplit('(', 1)[0] for mod in modules_avail]

    # Modules are used for finding model executable paths - so check
    # for unique module, or an exact match for the modulefile name
    if len(modules_avail) > 1 and modules_avail.count(modulefile) != 1:
        raise ValueError(
            f"There are multiple modules available for {modulefile}:\n" +
            f"{output}\n{MULTIPLE_MODULES_HELP}")
    elif len(modules_avail) == 0:
        raise ValueError(
            f"Module is not found: {modulefile}\n{MODULE_NOT_FOUND_HELP}"
        )


def run_module_cmd(subcommand, *args):
    """Wrapper around subprocess module command that captures output"""
    modulecmd = f"{os.environ['MODULESHOME']}/bin/modulecmd bash"
    command = f"{modulecmd} {subcommand} {' '.join(args)}"
    return subprocess.run(command, shell=True, text=True, capture_output=True)



"""payu.manifest
   ===============

   Provides an manifest class to store manifest data, which uses a
   subclassed yamanifest PayuManifest class

   :copyright: Copyright 2019 Aidan Heerdegen, see AUTHORS for details.
   :license: Apache License, Version 2.0, see LICENSE for details.
"""
from __future__ import print_function, absolute_import

# External
import fnmatch
import os
import sys
import shutil
import stat

from yamanifest.manifest import Manifest as YaManifest

from payu.fsops import make_symlink, mkdir_p


# fast_hashes = ['nchash','binhash']
fast_hashes = ['binhash']
full_hashes = ['md5']
all_hashes = fast_hashes + full_hashes


class PayuManifest(YaManifest):
    """
    A manifest object sub-classed from yamanifest object with some payu
    specific additions and enhancements.
    """
    def __init__(self, path,
                 ignore=None,
                 fast_hashes=fast_hashes,
                 full_hashes=full_hashes,
                 **kwargs):

        super(PayuManifest, self).__init__(path=path,
                                           hashes=fast_hashes+full_hashes,
                                           **kwargs)
        self.fast_hashes = fast_hashes
        self.full_hashes = full_hashes

        if ignore is not None:
            self.ignore = ignore

    def calculate_fast(self, previous_manifest):
        """
        Calculate hash value for all filepaths using a fast hash function and
        fall back to slower full hash functions if fast hashes fail to agree,
        with the pre-existing manifest
        """
        # Calculate all fast hashes
        self.add(
            filepaths=self.data.keys(),
            hashfn=self.fast_hashes,
            force=True,
            fullpaths=[self.fullpath(fpath) for fpath
                       in list(self.data.keys())]
        )

        # If fast hashes from previous manifest match, use previous full hashes
        # to avoid re-calculating slow hashes
        self.update_matching_hashes(other=previous_manifest)

        # Search for new files and files with changed fast hashes
        changed_filepaths = set()
        for filepath in self.data.keys():
            for hash in self.data[filepath]['hashes'].values():
                if hash is None:
                    changed_filepaths.add(filepath)

        # Calculate full hashes for these changed filepaths
        if len(changed_filepaths) > 0:
            self.add(
                filepaths=list(changed_filepaths),
                hashfn=self.full_hashes,
                force=True,
                fullpaths=[self.fullpath(fpath) for fpath
                           in list(changed_filepaths)]
            )

    def check_reproduce(self, previous_manifest):
        """
        Compare full hashes with previous manifest
        """
        # Use paths in both manifests to pick up new and missing files
        all_filepaths = set(self.data.keys()).union(
            previous_manifest.data.keys()
        )
        differences = []
        for filepath in all_filepaths:
            for hashfn in self.full_hashes:
                hash = self.get(filepath, hashfn)
                previous_hash = previous_manifest.get(filepath, hashfn)

                if hash is None:
                    differences.append(
                        f"  {filepath}: Missing file (file not in " +
                        "calculated manifest)"
                    )
                elif previous_hash is None:
                    differences.append(
                        f"  {filepath}: New file (file not in stored manifest)"
                    )
                elif hash != previous_hash:
                    differences.append(
                        f"  {filepath}: {hashfn}: {previous_hash} != {hash}"
                    )

        if len(differences) != 0:
            sys.stderr.write(
                f'Run cannot reproduce: manifest {self.path} is not correct\n'
            )
            print(f"Manifest path: stored hash != calculated hash")
            for row in differences:
                print(row)

            sys.exit(1)


    def add_filepath(self, filepath, fullpath, hashes, copy=False):
        """
        Bespoke function to add filepath & fullpath to manifest
        object without hashing. Can defer hashing until all files are
        added. Hashing all at once is much faster as overhead for
        threading is spread over all files
        """

        # Ignore directories
        if os.path.isdir(fullpath):
            return False

        # Ignore anything matching the ignore patterns
        for pattern in self.ignore:
            if fnmatch.fnmatch(os.path.basename(fullpath), pattern):
                return False

        if filepath not in self.data:
            self.data[filepath] = {}

        self.data[filepath]['fullpath'] = fullpath
        if 'hashes' not in self.data[filepath]:
            self.data[filepath]['hashes'] = {hash: None for hash in hashes}

        if copy:
            self.data[filepath]['copy'] = copy

        return True

    def add_fast(self, filepath, hashfn=None, force=False):
        """
        Bespoke function to add filepaths but set shortcircuit to True, which
        means only the first calculable hash will be stored. In this way only
        one "fast" hashing function need be called for each filepath.
        """
        if hashfn is None:
            hashfn = self.fast_hashes
        self.add(filepath, hashfn, force, shortcircuit=True)

    def copy_file(self, filepath):
        """
        Returns flag which says to copy rather than link a file.
        """
        copy_file = False
        try:
            copy_file = self.data[filepath]['copy']
        except KeyError:
            return False
        return copy_file

    def make_link(self, filepath):
        """
        Payu integration function for creating symlinks in work directories
        which point back to the original file.
        """
        if not os.path.exists(self.fullpath(filepath)):
            raise FileNotFoundError(
                "Unable to create symlink in work directory. "
                f"File not found: {self.fullpath(filepath)}"
            )
        else:
            try:
                destdir = os.path.dirname(filepath)
                # Make destination directory if not already exists
                # Necessary because sometimes this is called before
                # individual model setup
                if not os.path.exists(destdir):
                    os.makedirs(destdir)
                if self.copy_file(filepath):
                    shutil.copy(self.fullpath(filepath), filepath)
                    perm = (stat.S_IRUSR | stat.S_IRGRP
                            | stat.S_IROTH | stat.S_IWUSR)
                    os.chmod(filepath, perm)
                else:
                    make_symlink(self.fullpath(filepath), filepath)
            except Exception:
                action = 'copying' if self.copy_file else 'linking'
                print('payu: error: {action} orig: {orig} '
                      'local: {local}'.format(action=action,
                                              orig=self.fullpath(filepath),
                                              local=filepath))
                raise

    def copy(self, path):
        """
        Copy myself to another location
        """
        shutil.copy(self.path, path)

    def get_fullpaths(self):
        files = []
        for filepath in list(self):
            files.append(self.fullpath(filepath))
        return files

    def get_hashes(self, hashfn):
        hashes = []
        for filepath in self:
            hashes.append(self.get(filepath, hashfn))
        return hashes


class Manifest(object):
    """
    A Manifest class which stores all manifests for file tracking and
    methods to operate on them
    """

    def __init__(self, config, reproduce):

        # Manifest control configuration
        self.manifest_config = config

        # Not currently supporting specifying hash functions
        self.fast_hashes = self.manifest_config.get('fasthash', fast_hashes)
        self.full_hashes = self.manifest_config.get('fullhash', full_hashes)

        if type(self.fast_hashes) is str:
            self.fast_hashes = [self.fast_hashes, ]
        if type(self.full_hashes) is str:
            self.full_hashes = [self.full_hashes, ]

        self.ignore = self.manifest_config.get('ignore', ['.*'])
        if isinstance(self.ignore, str):
            self.ignore = [self.ignore]

        # Initialise manifests and reproduce flags
        self.manifests = {}
        self.previous_manifests = {}
        reproduce_config = self.manifest_config.get('reproduce', {})
        self.reproduce = {}
        for mf in ['input', 'restart', 'exe']:
            self.init_mf(mf)
            self.reproduce[mf] = reproduce_config.get(mf, reproduce)

        # Make sure the manifests directory exists
        mkdir_p(os.path.dirname(self.manifests['exe'].path))

    def init_mf(self, mf):
        # Initialise a sub-manifest object
        self.manifests[mf] = PayuManifest(
            os.path.join('manifests', '{}.yaml'.format(mf)),
            ignore=self.ignore,
            fast_hashes=self.fast_hashes,
            full_hashes=self.full_hashes
        )

        # Initialise a sub-manifest object to store pre-existing manifests
        self.previous_manifests[mf] = PayuManifest(
            os.path.join('manifests', '{}.yaml'.format(mf)),
            ignore=self.ignore,
            fast_hashes=self.fast_hashes,
            full_hashes=self.full_hashes
        )

    def __iter__(self):
        """
        Iterator method
        """
        for mf in self.manifests:
            yield self.manifests[mf]

    def __len__(self):
        """Return the number of manifests in the manifest class."""
        return len(self.manifests)

    def load_manifests(self):
        """
        Load pre-existing manifests
        """
        for mf in self.previous_manifests:
            manifest_path = self.previous_manifests[mf].path
            if os.path.exists(manifest_path):
                try:
                    print(f'Loading {mf} manifest: {manifest_path}')
                    self.previous_manifests[mf].load()
                except Exception as e:
                    print(f'Error loading {mf} manifest: {e}')

            # Check manifests are populated when reproduce is configured
            if len(self.previous_manifests[mf]) == 0 and self.reproduce[mf]:
                sys.stderr.write(
                    f'{mf.capitalize()} manifest must exist and be populated '
                    'if reproduce is configured to True\n'
                )
                sys.exit(1)

    def setup(self):
        # Load all available manifests
        self.load_manifests()

    def check_manifests(self):
        print("Checking exe, input and restart manifests")
        for mf in self.manifests:
            # Calculate hashes in manifests
            self.manifests[mf].calculate_fast(self.previous_manifests[mf])

            if self.reproduce[mf]:
                # Compare manifest with previous manifest
                self.manifests[mf].check_reproduce(self.previous_manifests[mf])

        # Update manifests if there's any changes, or create file if empty
        for mf in self.manifests:
            if (self.manifests[mf].data != self.previous_manifests[mf].data
                    or len(self.manifests[mf]) == 0):
                print("Writing {}".format(self.manifests[mf].path))
                self.manifests[mf].dump()

    def copy_manifests(self, path):

        mkdir_p(path)
        try:
            for mf in self.manifests:
                self.manifests[mf].copy(path)
        except IOError:
            pass

    def add_filepath(self, manifest, filepath, fullpath, copy=False):
        """
        Wrapper to the add_filepath function in PayuManifest. Prevents outside
        code from directly calling anything in PayuManifest.
        """
        filepath = os.path.normpath(filepath)
        if self.manifests[manifest].add_filepath(
                filepath=filepath,
                fullpath=fullpath,
                hashes=self.fast_hashes + self.full_hashes,
                copy=copy):
            # Only link if filepath was added
            self.manifests[manifest].make_link(filepath)

    def get_all_previous_fullpaths(self):
        """
        Return a list of all fullpaths in manifest files
        """
        files = []
        for mf in self.previous_manifests:
            files.extend(self.previous_manifests[mf].get_fullpaths())
        return files



"""Payu experiment UUID and metadata support

Generates and commit a new experiment uuid and updates/creates experiment
metadata

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

import requests
import shutil
import uuid
import warnings
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from payu.fsops import read_config, mkdir_p
from payu.git_utils import GitRepository

# A truncated uuid is used for branch-uuid aware experiment names
TRUNCATED_UUID_LENGTH = 8

# Metadata file field names
UUID_FIELD = "experiment_uuid"
PARENT_UUID_FIELD = "parent_experiment"
CONTACT_FIELD = "contact"
EMAIL_FIELD = "email"
NAME_FIELD = "name"
GIT_URL_FIELD = "url"
CREATED_FIELD = "created"
MODEL_FIELD = "model"
METADATA_FILENAME = "metadata.yaml"

# Metadata Schema
SCHEMA_URL = "https://raw.githubusercontent.com/ACCESS-NRI/schema/80a3ce720af14b2b5e718630e1b52e7b3d22ea95/au.org.access-nri/model/output/experiment-metadata/1-0-3.json"

class MetadataWarning(Warning):
    pass


class Metadata:
    """
    Class to store/update/create metadata such as experiment uuid and name

    Parameters:
        laboratory_archive_path : Path
            The archive sub-directory in Laboratory
        branch : Optional[str]
            The git branch on which the experiment is run
        control_path : Optional[Path]
            Path to where the experiment is configured and run. The default
            is set to the current working directory. This default is set in
            in fsops.read_config
        config_path : Optional[Path]
            Configuration Path. The default is config.yaml in the current
            working directory. This is also set in fsop.read_config
        disabled : bool, default False
            Flag to disable metadata and UUID generation and commits. The
            legacy name (control directory name) for experiments names
            in archive will be used instead.
    """

    def __init__(self,
                 laboratory_archive_path: Path,
                 config_path: Optional[Path] = None,
                 branch: Optional[str] = None,
                 control_path: Optional[Path] = None,
                 disabled: Optional[bool] = False) -> None:
        self.config = read_config(config_path)
        self.metadata_config = self.config.get('metadata', {})

        if control_path is None:
            control_path = Path(self.config.get("control_path"))
        self.control_path = control_path
        self.filepath = self.control_path / METADATA_FILENAME
        self.lab_archive_path = laboratory_archive_path

        # Check if metadata has been disabled in call, env flag under PBS,
        # or in config.yaml
        self.enabled = (
            not disabled and
            self.metadata_config.get('enable', True)
        )

        if self.enabled:
            self.repo = GitRepository(self.control_path, catch_error=True)

        self.branch = branch

        # Set uuid if in metadata file
        metadata = self.read_file()
        self.uuid = metadata.get(UUID_FIELD, None)
        self.uuid_updated = False

    def read_file(self) -> CommentedMap:
        """Read metadata file - preserving orginal format if it exists"""
        metadata = CommentedMap()
        if self.filepath.exists():
            # Use default ruamel YAML to preserve comments and multi-line
            # strings
            metadata = YAML().load(self.filepath)
        return metadata

    def setup(self,
              is_new_experiment: bool = False,
              keep_uuid: bool = False) -> None:
        """Set UUID and experiment name.

        Parameters:
            keep_uuid: bool, default False
                Keep pre-existing UUID, if it exists.
            is_new_experiment: bool, default False
                If not keep_uuid, generate a new UUID and a branch-uuid aware
                experiment name. This is set in payu.branch.checkout_branch.
        Return: None

        Note: Experiment name is the name used for the work and archive
        directories in the Laboratory.
        """
        if not self.enabled:
            # Set experiment name only - either configured or legacy name
            self.set_experiment_name()

        elif self.uuid is not None and (keep_uuid or not is_new_experiment):
            self.set_experiment_name(keep_uuid=keep_uuid,
                                     is_new_experiment=is_new_experiment)
        else:
            # Generate new UUID
            if self.uuid is None and not is_new_experiment:
                warnings.warn("No experiment uuid found in metadata. "
                              "Generating a new uuid", MetadataWarning)
            self.set_new_uuid(is_new_experiment=is_new_experiment)

        self.archive_path = self.lab_archive_path / self.experiment_name

    def new_experiment_name(self) -> str:
        """Generate a new experiment name"""
        if self.branch is None:
            self.branch = self.repo.get_branch_name()

        # Add branch and a truncated uuid to control directory name
        adding_branch = self.branch not in (None, 'main', 'master')
        suffix = f'-{self.branch}' if adding_branch else ''

        truncated_uuid = self.uuid[:TRUNCATED_UUID_LENGTH]
        suffix += f'-{truncated_uuid}'

        return self.control_path.name + suffix

    def set_experiment_name(self,
                            is_new_experiment: bool = False,
                            keep_uuid: bool = False) -> None:
        """Set experiment name - this is used for work and archive
        sub-directories in the Laboratory"""
        # Experiment name configuration - this overrides experiment name
        self.experiment_name = self.config.get("experiment", None)
        if self.experiment_name is not None:
            print(f"Experiment name is configured in config.yaml: ",
                  self.experiment_name)
            return

        # Legacy experiment name
        legacy_name = self.control_path.name

        if not self.enabled:
            # Metadata/UUID generation is disabled, so leave UUID out of
            # experiment name
            self.experiment_name = legacy_name
            print("Metadata and UUID generation is disabled.",
                  f"Experiment name used for archival: {self.experiment_name}")
            return

        branch_uuid_experiment_name = self.new_experiment_name()
        if is_new_experiment or self.has_archive(branch_uuid_experiment_name):
            # Use branch-UUID aware experiment name
            self.experiment_name = branch_uuid_experiment_name
        elif self.has_archive(legacy_name):
            # Use legacy CONTROL-DIR experiment name
            self.experiment_name = legacy_name
        elif keep_uuid:
            # Use same experiment UUID and use branch-UUID name for archive
            self.experiment_name = branch_uuid_experiment_name
        else:
            # No archive exists - Detecting new experiment
            warnings.warn(
                "No pre-existing archive found. Generating a new uuid",
                MetadataWarning
            )
            self.set_new_uuid(is_new_experiment=True)

    def has_archive(self, experiment_name: str) -> bool:
        """Return True if archive under the experiment name exists and
        if it exists, check for a non-matching UUID in archive metadata."""
        archive_path = self.lab_archive_path / experiment_name

        if archive_path.exists():
            # Check if the UUID in the archive metadata matches the
            # UUID in metadata
            archive_metadata_path = archive_path / METADATA_FILENAME
            if archive_metadata_path.exists():
                archive_metadata = YAML().load(archive_metadata_path)
                if (UUID_FIELD in archive_metadata and
                        archive_metadata[UUID_FIELD] != self.uuid):
                    print("Mismatch of UUIDs between metadata and an archive "
                          f"metadata found at: {archive_metadata_path}")
                    return False
            print(f"Found experiment archive: {archive_path}")
        return archive_path.exists()

    def set_new_uuid(self, is_new_experiment: bool = False) -> None:
        """Generate a new uuid and set experiment name"""
        self.uuid_updated = True
        self.uuid = generate_uuid()
        self.set_experiment_name(is_new_experiment=is_new_experiment)

        # If experiment name does not include UUID, leave it unchanged
        if self.experiment_name.endswith(self.uuid[:TRUNCATED_UUID_LENGTH]):
            return

        # Check experiment name is unique in local archive
        lab_archive_path = self.lab_archive_path
        if lab_archive_path.exists():
            local_experiments = [item for item in lab_archive_path.iterdir()
                                 if item.is_dir()]
            while self.experiment_name in local_experiments:
                # Generate a new id and experiment name
                self.uuid = generate_uuid()
                self.set_experiment_name(is_new_experiment=is_new_experiment)

    def write_metadata(self,
                       restart_path: Optional[Union[Path, str]] = None,
                       set_template_values: bool = False,
                       parent_experiment: Optional[str] = None) -> None:
        """Create/update metadata file, commit any changes and
        copy metadata file to the experiment archive.

        Parameters:
            restart_path: Optional[Path]
                Prior restart path - used for finding parent metadata
            set_template_values: bool, default False
                Read schema and set metadata template values for new
                experiments
            parent_experiment: Optional[str]
                Parent experiment UUID to add to generated metadata

        Return: None

        Note: This assumes setup() has been run to set UUID and experiment name
        """
        if not self.enabled:
            # Skip creating/updating/commiting metadata
            return

        if self.uuid_updated:
            # Update metadata if UUID has changed
            restart_path = Path(restart_path) if restart_path else None
            self.update_file(restart_path=restart_path,
                             set_template_values=set_template_values,
                             parent_experiment=parent_experiment)
            self.commit_file()

        self.copy_to_archive()

    def update_file(self,
                    restart_path: Optional[Path] = None,
                    set_template_values: bool = False,
                    parent_experiment: Optional[str] = None) -> None:
        """Write any updates to metadata file"""
        metadata = self.read_file()

        # Add UUID field
        metadata[UUID_FIELD] = self.uuid

        # Update parent UUID field
        if parent_experiment is None:
            parent_experiment = self.get_parent_experiment(restart_path)
        if parent_experiment and parent_experiment != self.uuid:
            metadata[PARENT_UUID_FIELD] = parent_experiment

        # Add extra fields if new branch-uuid experiment
        # so to not over-write fields if it's a pre-existing legacy experiment
        if self.experiment_name.endswith(self.uuid[:TRUNCATED_UUID_LENGTH]):
            metadata[CREATED_FIELD] = datetime.now().strftime('%Y-%m-%d')
            metadata[NAME_FIELD] = self.experiment_name
            metadata[MODEL_FIELD] = self.get_model_name()

            # Add origin git URL, if defined
            url = self.repo.get_origin_url()
            if url:
                metadata[GIT_URL_FIELD] = url

            # Add email + contact if defined in git config
            contact = self.repo.get_user_info(config_key='name')
            if contact:
                metadata[CONTACT_FIELD] = contact

            email = self.repo.get_user_info(config_key="email")
            if email:
                metadata[EMAIL_FIELD] = email

        if set_template_values:
            # Note that retrieving schema requires internet access
            add_template_metadata_values(metadata)

        # Write updated metadata to file
        YAML().dump(metadata, self.filepath)

    def get_model_name(self) -> str:
        """Get model name from config file"""
        # Use capitilised model name unless a specific model name is defined
        default_model_name = self.config.get('model').upper()
        model_name = self.metadata_config.get('model', default_model_name)
        return model_name

    def get_parent_experiment(self, prior_restart_path: Path) -> None:
        """Searches UUID in the metadata in the parent directory that
        contains the restart"""
        if prior_restart_path is None:
            return

        # Resolve to absolute path
        prior_restart_path = prior_restart_path.resolve()

        # Check for pre-existing metadata file
        base_output_directory = Path(prior_restart_path).parent
        metadata_filepath = base_output_directory / METADATA_FILENAME
        if not metadata_filepath.exists():
            return

        # Read metadata file
        parent_metadata = YAML().load(metadata_filepath)
        return parent_metadata.get(UUID_FIELD, None)

    def commit_file(self) -> None:
        """Add a git commit for changes to metadata file, if file has changed
        and if control path is a git repository"""
        commit_message = f"Updated metadata. Experiment UUID: {self.uuid}"
        self.repo.commit(commit_message=commit_message,
                         paths_to_commit=[self.filepath])

    def copy_to_archive(self) -> None:
        """Copy metadata file to archive"""
        mkdir_p(self.archive_path)
        shutil.copy(self.filepath, self.archive_path / METADATA_FILENAME)
        # Note: The existence of an archive is used for determining
        # experiment names and whether to generate a new UUID


def get_schema_from_github():
    """Retrieve metadata schema from github"""
    response = requests.get(SCHEMA_URL)

    if response.status_code == 200:
        return response.json()
    else:
        print(f"Failed to fetch schema from {SCHEMA_URL}")
    return response.json() if response.status_code == 200 else {}


def add_template_metadata_values(metadata: CommentedMap) -> None:
    """Add in templates for un-set metadata values"""
    schema = get_schema_from_github()

    for key, value in schema.get('properties', {}).items():
        if key not in metadata:
            # Add field with commented description of value
            description = value.get('description', None)
            if description is not None:
                metadata[key] = None
                metadata.yaml_add_eol_comment(description, key)


def generate_uuid() -> str:
    """Generate a new uuid"""
    return str(uuid.uuid4())



"""Driver interface to the MOM ocean model.

:copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
:license: Apache License, Version 2.0, see LICENSE for details
"""
import os
import shlex
import shutil
import subprocess

import f90nml

from payu.models.fms import Fms
from payu.models.mom_mixin import MomMixin
from payu.fsops import mkdir_p, make_symlink


class Mom(MomMixin, Fms):

    def __init__(self, expt, name, config):

        # FMS initialisation
        super(Mom, self).__init__(expt, name, config)

        # Model-specific configuration
        self.model_type = 'mom'
        self.default_exec = 'fms_MOM_SIS.x'

        # Default repo and build details.
        self.repo_url = 'git://github.com/BreakawayLabs/mom.git'
        self.repo_tag = 'master'
        self.build_command = './MOM_compile.csh --platform nci --type MOM_SIS'

        self.config_files = [
            'data_table',
            'diag_table',
            'field_table',
            'input.nml'
        ]

        self.optional_config_files = [
            'blob_diag_table',
            'mask_table',
            'ocean_mask_table'
        ]

    def set_model_pathnames(self):
        super(Mom, self).set_model_pathnames()

        self.build_exec_path = os.path.join(self.codebase_path, 'exec', 'nci',
                                            'MOM_SIS')
        self.build_path = os.path.join(self.codebase_path, 'exp')

    def build_model(self):
        super(Mom, self).build_model()

        # Model is built, now copy over mppnccombine.
        mppnc_exec = 'mppnccombine.nci'

        mppnc_src = os.path.join(self.codebase_path, 'bin', mppnc_exec)
        mppnc_dest = os.path.join(self.expt.lab.bin_path, 'mppnccombine')
        shutil.copy(mppnc_src, mppnc_dest)

    def setup(self):
        # FMS initialisation
        super(Mom, self).setup()

        if not self.top_level_model:
            # Make log dir
            mkdir_p(os.path.join(self.work_path, 'log'))

        input_nml_path = os.path.join(self.work_path, 'input.nml')
        input_nml = f90nml.read(input_nml_path)

        # Set the runtime
        if self.expt.runtime:
            ocean_solo_nml = input_nml['ocean_solo_nml']

            ocean_solo_nml['years'] = self.expt.runtime['years']
            ocean_solo_nml['months'] = self.expt.runtime['months']
            ocean_solo_nml['days'] = self.expt.runtime['days']
            ocean_solo_nml['seconds'] = self.expt.runtime.get('seconds', 0)

            input_nml.write(input_nml_path, force=True)

        # Construct the land CPU mask
        if self.expt.config.get('mask_table', False):
            # NOTE: This function actually creates a mask table using the
            #       `check_mask` command line tool.  But it is not very usable
            #       since you need to know the number of masked CPUs to submit
            #       the job.  It needs a rethink of the submission process.
            self.create_mask_table(input_nml)

        # NOTE: Don't expect this to be here forever...
        # Attempt to set a mask table from the input
        if self.config.get('mask', False):
            mask_path = os.path.join(self.work_input_path, 'ocean_mask_table')

            # Remove any existing mask
            # (If no reference mask is available, then we will not use one)
            if os.path.isfile(mask_path):
                os.remove(mask_path)

            # Reference mask table
            assert ('layout' in input_nml['ocean_model_nml'])
            nx, ny = input_nml['ocean_model_nml'].get('layout')
            n_masked_cpus = nx * ny - self.config.get('ncpus')

            mask_table_fname = 'mask_table.{nmask}.{nx}x{ny}'.format(
                nmask=n_masked_cpus,
                nx=nx,
                ny=ny
            )

            ref_mask_path = os.path.join(self.work_input_path,
                                         mask_table_fname)

            # Set (or replace) mask table if reference is available
            if os.path.isfile(ref_mask_path):
                make_symlink(ref_mask_path, mask_path)

    def set_timestep(self, timestep):

        input_nml_path = os.path.join(self.work_path, 'input.nml')
        input_nml = f90nml.read(input_nml_path)

        input_nml['ocean_model_nml']['dt_ocean'] = timestep

        input_nml.write(input_nml_path, force=True)

    def create_mask_table(self, input_nml):
        import netCDF4

        # Disable E1136 which is tripped below when accessing grid_vars
        # pylint: disable=unsubscriptable-object

        # Get the grid spec path
        grid_spec_fname = 'grid_spec.nc'
        for input_dir in self.input_paths:
            if grid_spec_fname in os.listdir(input_dir):
                grid_spec_path = os.path.join(input_dir, grid_spec_fname)
                break
        assert grid_spec_path

        grid_spec_nc = netCDF4.Dataset(grid_spec_path)
        grid_vars = grid_spec_nc.variables

        # Get the ocean mosaic file
        # TODO: Do not assume mosaic format
        ocn_mosaic_fname = ''.join(grid_vars['ocn_mosaic_file'][:].data)
        for input_dir in self.input_paths:
            if ocn_mosaic_fname in os.listdir(input_dir):
                ocn_mosaic_path = os.path.join(input_dir, ocn_mosaic_fname)
                break

        # Get the topography file
        ocn_topog_fname = ''.join(grid_vars['ocn_topog_file'][:].data)
        for input_dir in self.input_paths:
            if ocn_topog_fname in os.listdir(input_dir):
                ocn_topog_path = os.path.join(input_dir, ocn_topog_fname)
                break

        # pylint: enable=unsubscriptable-object

        grid_spec_nc.close()

        check_mask = os.path.join(self.expt.lab.bin_path, 'check_mask')
        f_null = open(os.devnull, 'w')

        # Generate ocean mask_table
        ocn_layout = input_nml['ocean_model_nml']['layout']

        cmd = (
            '{check_mask} --grid_file {grid_file} '
            '--ocean_topog {ocean_topog} --layout {layout}'.format(
                check_mask=check_mask,
                grid_file=ocn_mosaic_path,
                ocean_topog=ocn_topog_path,
                layout=','.join([str(s) for s in ocn_layout])
            )
        )
        subprocess.call(shlex.split(cmd), stdout=f_null)
        ocn_mask_fname = [f for f in os.listdir(os.curdir)
                          if f.startswith('mask_table')][0]

        ocn_mask_path = os.path.join(self.work_input_path,
                                     'ocean_mask_table')
        shutil.copy(ocn_mask_fname, ocn_mask_path)

        # Generate the ice mask_table
        ice_layout = input_nml['ice_model_nml']['layout']

        if ice_layout == ocn_layout:
            ice_mask_fname = ocn_mask_fname
        else:
            cmd = (
                '{check_mask} --grid_file {grid_file} '
                '--ocean_topog {ocean_topog} --layout {layout}'.format(
                    check_mask=check_mask,
                    grid_file=ocn_mosaic_path,
                    ocean_topog=ocn_topog_path,
                    layout=','.join([str(s) for s in ice_layout])
                )
            )
            subprocess.call(shlex.split(cmd), stdout=f_null)
            ice_mask_fname = [f for f in os.listdir(os.curdir)
                              if f.startswith('mask_table')][0]

        ice_mask_path = os.path.join(self.work_input_path,
                                     'ice_mask_table')

        shutil.copy(ice_mask_fname, ice_mask_path)

        try:
            os.remove(ocn_mask_fname)
            os.remove(ice_mask_fname)
        except EnvironmentError:
            # TODO: Check this a little bit better
            pass

        f_null.close()

        # Read and return the number of land cells
        with open(ocn_mask_path) as fmask:
            land_cells = int(fmask.readline())

        return land_cells


def get_restart_datetime_using_mom_submodel(model, restart_path):
    """Given a model object and a restart path, check for
    a MOM sub-model and use it's restart files to find a cftime datetime 
   (for date-based restart pruning)"""
    for sub_model in model.expt.models:
        if sub_model.model_type == 'mom' :
            mom_restart_path = os.path.join(restart_path, sub_model.name)
            return sub_model.get_restart_datetime(mom_restart_path)

    raise NotImplementedError(
        f'Cannot find mom sub-model: {model.model_type} date-based '
        'restart pruning requires the mom sub-model to '
        'determine restart dates'
    )


"""payu.models.um
   ==============

   The payu interface for the UM atmosphere model

   :copyright: Copyright 2011 Marshall Ward, see AUTHORS for details
   :license: Apache License, Version 2.0, see LICENSE for details
"""
from __future__ import print_function

# Standard Library
import datetime
import fileinput
import glob
import os
import shutil
import string

# Extensions
import f90nml
import yaml

# Local
from payu.fsops import mkdir_p, make_symlink
from payu.models.model import Model
import payu.calendar as cal


class UnifiedModel(Model):

    def __init__(self, expt, name, config):
        super(UnifiedModel, self).__init__(expt, name, config)

        self.model_type = 'um'
        self.default_exec = 'um'
        self.restart_calendar_file = self.model_type + '.res.yaml'

        # TODO: many of these can probably be ignored.
        self.config_files = [
            'errflag',
            'hnlist', 'ihist',
            'namelists', 'prefix.PRESM_A',
            'STASHC', 'UAFILES_A', 'UAFLDS_A',
            'cable.nml', 'um_env.yaml'
            ]
        self.optional_config_files = ['input_atm.nml', 'parexe']

        self.restart = 'restart_dump.astart'

    def set_model_pathnames(self):
        super(UnifiedModel, self).set_model_pathnames()

        self.work_input_path = os.path.join(self.work_path, 'INPUT')

    def archive(self):
        super(UnifiedModel, self).archive()

        # Delete all the stdout log files except the root PE
        # Sorts to ensure root PE is first entry
        files = sorted(glob.glob(
                       os.path.join(self.work_path, 'atm.fort6.pe*')),
                       key=lambda name: int(name.rpartition('.')[-1][2:]))
        if len(files) > 1:
            for f_path in files[1:]:
                os.remove(f_path)

        mkdir_p(self.restart_path)

        # Need to figure out the end date of the model.
        nml_path = os.path.join(self.work_path, 'namelists')
        nml = f90nml.read(nml_path)

        resubmit_inc = nml['NLSTCALL']['RUN_RESUBMIT_INC']
        runtime = um_time_to_time(resubmit_inc)
        # runtime = datetime.timedelta(seconds=runtime)

        basis_time = nml['NLSTCALL']['MODEL_BASIS_TIME']
        init_date = um_date_to_date(basis_time)

        end_date = cal.date_plus_seconds(init_date,
                                         runtime,
                                         cal.GREGORIAN)

        # Save model time to restart next run
        with open(os.path.join(self.restart_path,
                  self.restart_calendar_file), 'w') as restart_file:
            restart_file.write(yaml.dump({'end_date': end_date},
                               default_flow_style=False))

        end_date = date_to_um_dump_date(end_date)

        restart_dump = os.path.join(self.work_path,
                                    'aiihca.da{0}'.format(end_date))
        f_dst = os.path.join(self.restart_path, self.restart)
        if os.path.exists(restart_dump):
            shutil.copy(restart_dump, f_dst)
        else:
            print('payu: error: Model has not produced a restart dump file:\n'
                  '{} does not exist.\n'
                  'Check DUMPFREQim in namelists'.format(restart_dump))

        # Now remove restart files from work directory so they're not
        # unnecessarily archived to output
        for f_path in glob.glob(os.path.join(self.work_path, 'aiihca.da*')):
            os.remove(f_path)

    def collate(self):
        pass

    def setup(self):
        # Raise a deprecation error if the um_env.yaml file is not found
        # This could be removed down the line, once older configurations 
        # have swapped to um_env.yaml files.
        deprecated_um_env = os.path.join(self.control_path, 'um_env.py')
        new_um_env = os.path.join(self.control_path, 'um_env.yaml')
        if (not os.path.isfile(new_um_env)) and os.path.isfile(deprecated_um_env):
            raise RuntimeError(
                (
                    "The `um_env.py` configuration file is no longer "
                    "supported and should be replaced with a yaml file. "
                    "Convert `um_env.py` to `um_env.yaml` using "
                    "https://github.com/ACCESS-NRI/esm1.5-scripts/blob/main/config-files/UM/um_env_to_yaml.py"
                )
            ) 

        # Commence normal setup
        super(UnifiedModel, self).setup()

        # Set up environment variables needed to run UM.
        um_env_path = os.path.join(self.control_path, 'um_env.yaml')
        with open(um_env_path, 'r') as um_env_yaml:
            um_env_vars = yaml.safe_load(um_env_yaml)


        # Stage the UM restart file.
        if self.prior_restart_path and not self.expt.repeat_run:
            f_src = os.path.join(self.prior_restart_path, self.restart)
            f_dst = os.path.join(self.work_input_path, self.restart)

            if os.path.isfile(f_src):
                make_symlink(f_src, f_dst)
                # every run is an NRUN with an updated ASTART file
                um_env_vars['ASTART'] = self.restart
                um_env_vars['TYPE'] = 'NRUN'

        # Set paths in environment variables.
        for k in um_env_vars.keys():
            um_env_vars[k] = um_env_vars[k].format(
                                    input_path=self.input_paths[0],
                                    work_path=self.work_path
            )
        os.environ.update(um_env_vars)

        # parexe removed from newer configurations - retain the
        # old processing if parexe file exists for backwards compatibility
        parexe = os.path.join(self.work_path, 'parexe')
        if os.path.isfile(parexe):
            # The above needs to be done in parexe also.
            # FIXME: a better way to do this or remove.
            for line in fileinput.input(parexe, inplace=True):
                line = line.format(input_path=self.input_paths[0],
                                   work_path=self.work_path)
                print(line, end='')


        work_nml_path = os.path.join(self.work_path, 'namelists')
        work_nml = f90nml.read(work_nml_path)

        restart_calendar_path = os.path.join(self.work_init_path,
                                             self.restart_calendar_file)

        # Modify namelists for a continuation run.
        if self.prior_restart_path and not self.expt.repeat_run \
           and os.path.exists(restart_calendar_path):

            with open(restart_calendar_path, 'r') as restart_file:
                restart_info = yaml.safe_load(restart_file)

            run_start_date = restart_info['end_date']

            # Write out and save new calendar information.
            run_start_date_um = date_to_um_date(run_start_date)
            work_nml['NLSTCALL']['MODEL_BASIS_TIME'] = run_start_date_um
            work_nml['NLSTCALL']['ANCIL_REFTIME'] = run_start_date_um

        else:
            run_start_date = work_nml['NLSTCALL']['MODEL_BASIS_TIME']
            run_start_date = um_date_to_date(run_start_date)

        # Set the runtime for this run.
        if self.expt.runtime:
            run_runtime = cal.runtime_from_date(
                run_start_date,
                self.expt.runtime['years'],
                self.expt.runtime['months'],
                self.expt.runtime['days'],
                self.expt.runtime.get('seconds', 0),
                cal.GREGORIAN)
            run_runtime = time_to_um_time(run_runtime)
            work_nml['NLSTCALL']['RUN_RESUBMIT_INC'] = run_runtime
            work_nml['NLSTCALL']['RUN_TARGET_END'] = run_runtime
            work_nml['STSHCOMP']['RUN_TARGET_END'] = run_runtime

        work_nml.write(work_nml_path, force=True)


def date_to_um_dump_date(date):
    """
    Convert a time date object to a um dump format date which is
    <decade><year><month><day>0

    To accommodate two digit months and days the UM uses letters. e.g. 1st oct
    is writing 01a10.
    """

    assert(date.month <= 12)

    decade = date.year // 10
    # UM can only handle 36 decades then goes back to the beginning.
    decade = decade % 36
    year = date.year % 10
    month = date.month
    day = date.day

    um_d = string.digits + string.ascii_letters[:26]

    um_dump_date = (
        '{decade}{year}{month}{day}0'.format(
            decade=um_d[decade],
            year=um_d[year],
            month=um_d[month],
            day=um_d[day]
        )
    )
    return um_dump_date


def date_to_um_date(date):
    """
    Convert a date object to 'year, month, day, hour, minute, second.'
    """

    assert date.hour == 0 and date.minute == 0 and date.second == 0

    return [date.year, date.month, date.day, 0, 0, 0]


def um_date_to_date(d):
    """
    Convert a string with format 'year, month, day, hour, minute, second'
    to a datetime date.
    """

    return datetime.datetime(year=d[0], month=d[1], day=d[2],
                             hour=d[3], minute=d[4], second=d[5])


def um_time_to_time(d):
    """
    Convert a list with format [year, month, day, hour, minute, second]
    to a number of seconds.

    Only days are supported.
    """

    assert d[0] == 0 and d[1] == 0 and d[3] == 0 and d[4] == 0 and d[5] == 0

    return d[2]*86400


def time_to_um_time(seconds):
    """
    Convert a number of seconds to a list with format [year, month, day, hour,
       minute, second]

    Only days are supported.
    """

    assert(seconds % 86400 == 0)

    return [0, 0, seconds // 86400, 0, 0, 0]



"""Run the `payu branch` command.

:copyright: Copyright 2018 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""

from pathlib import Path

from payu.branch import list_branches
import payu.subcommands.args as args

title = 'branch'
parameters = {'description': ('List git branches and corresponding metadata')}

arguments = [args.config, args.verbose, args.remote]


def runcmd(config_path, verbose, remote):
    """Execute the command."""
    config_path = Path(config_path) if config_path is not None else None
    list_branches(config_path, verbose, remote)


runscript = runcmd



import pytest
from unittest.mock import patch

from payu.envmod import check_modulefile


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_unique(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0
"""
    # Test runs without an error
    check_modulefile('test-module/1.0.0')


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_without_version(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0
test-module/2.0.0
test-module/3.0.1
"""

    # Expect an error raised
    with pytest.raises(ValueError) as exc_info:
        check_modulefile('test-module')
        exc_info.value.startswith(
            "There are multiple modules available for test-module"
        )

    # Mock module avail command use debug in name
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0
test-module/1.0.0-debug
"""

    # Expect an error raised
    with pytest.raises(ValueError) as exc_info:
        check_modulefile('test-module')
        exc_info.value.startswith(
            "There are multiple modules available for test-module"
        )


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_exact_match(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0
test-module/1.0.0-debug
"""

    # Test runs without an error
    check_modulefile('test-module/1.0.0')


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_exact_match_with_symbolic_version(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0(default)
test-module/1.0.0-debug
"""

    # Test runs without an error
    check_modulefile('test-module/1.0.0')

    # Rerun test with another symbolic version/alias other than default
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0(some_symbolic_name_or_alias)
test-module/1.0.0-debug
"""

    # Test runs without an error
    check_modulefile('test-module/1.0.0')


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_multiple_modules(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = """/path/to/modulefiles:
test-module/1.0.0
/another/module/path:
test-module/1.0.0
"""

    # Expect an error raised
    with pytest.raises(ValueError) as exc_info:
        check_modulefile('test-module/1.0.0')
        exc_info.value.startswith(
            "There are multiple modules available for test-module"
        )


@patch('payu.envmod.run_module_cmd')
def test_check_modulefile_no_modules_found(mock_module_cmd):
    # Mock module avail command
    mock_module_cmd.return_value.stderr = ""

    # Expect an error raised
    with pytest.raises(ValueError) as exc_info:
        check_modulefile('test-module/1.0.0')
        exc_info.value.startswith("Module is not found: test-module")



import cftime
import datetime
import pytest

from payu.calendar import parse_date_offset, DatetimeOffset
from payu.calendar import seconds_between_dates, int_to_date
from payu.calendar import runtime_from_date
from payu.calendar import GREGORIAN, NOLEAP

SEC_PER_DAY = 24*60*60


@pytest.mark.parametrize(
    "offset, initial_dt, expected",
    [
        (
            "5YS",
            cftime.datetime(year=2000, month=10, day=31,
                            hour=10, minute=5, second=2,
                            calendar="noleap"),
            cftime.datetime(year=2005, month=1, day=1, calendar="noleap"),
        ),
        (
            "1YS",
            cftime.datetime(year=1700, month=2, day=15,
                            hour=11, minute=5, second=2,
                            calendar="proleptic_gregorian"),
            cftime.datetime(year=1701, month=1, day=1,
                            calendar="proleptic_gregorian"),
        ),
        (
            "20YS",
            cftime.datetime(year=2200, month=2, day=30,
                            hour=1, minute=4, second=20,
                            calendar="360_day"),
            cftime.datetime(year=2220, month=1, day=1, calendar="360_day"),
        ),
    ],
)
def test_year_start_date_offset_add_to_datetime(offset, initial_dt, expected):
    date_offset = parse_date_offset(offset)
    next_dt = date_offset.add_to_datetime(initial_dt)

    assert next_dt == expected


@pytest.mark.parametrize(
    "offset, initial_dt, expected",
    [
        (
            "5MS",
            cftime.datetime(year=2000, month=10, day=1,
                            hour=10, minute=5, second=2,
                            calendar="noleap"),
            cftime.datetime(year=2001, month=3, day=1, calendar="noleap"),
        ),
        (
            "13MS",
            cftime.datetime(year=1500, month=10, day=30,
                            hour=10, minute=5, second=2,
                            calendar="360_day"),
            cftime.datetime(year=1501, month=11, day=1, calendar="360_day"),
        ),
        (
            "24MS",
            cftime.datetime(year=2200, month=1, day=1, calendar="gregorian"),
            cftime.datetime(year=2202, month=1, day=1, calendar="gregorian"),
        ),
    ],
)
def test_month_start_date_offset_add_to_datetime(offset, initial_dt, expected):
    date_offset = parse_date_offset(offset)
    next_dt = date_offset.add_to_datetime(initial_dt)

    assert next_dt == expected


@pytest.mark.parametrize(
    "offset, initial_dt, expected",
    [
        (
            "100S",
            cftime.datetime(year=2000, month=10, day=31,
                            hour=10, minute=5, second=2,
                            calendar="julian"),
            cftime.datetime(year=2000, month=10, day=31,
                            hour=10, minute=6, second=42,
                            calendar="julian"),
        ),
        (
            "25H",
            cftime.datetime(year=1500, month=10, day=30,
                            hour=10, minute=5, second=2,
                            calendar="360_day"),
            cftime.datetime(year=1500, month=11, day=1,
                            hour=11, minute=5, second=2,
                            calendar="360_day")
        ),
        (
            "3W",
            cftime.datetime(year=2200, month=1, day=1),
            cftime.datetime(year=2200, month=1, day=22),
        ),
        (
            "4T",
            cftime.datetime(
                year=2200, month=1, day=1, hour=0, minute=0, second=0
            ),
            cftime.datetime(
                year=2200, month=1, day=1, hour=0, minute=4, second=0
            ),
        ),
        (
            "30D",
            cftime.datetime(year=2200, month=2, day=1,  calendar="noleap"),
            cftime.datetime(year=2200, month=3, day=3, calendar="noleap"),
        ),
    ],
)
def test_timedelta_date_offset_add_to_datetime(offset, initial_dt, expected):
    # Week, Day, Minute, Hour, Second offsets
    date_offset = parse_date_offset(offset)
    next_dt = date_offset.add_to_datetime(initial_dt)

    assert next_dt == expected


def test_date_offset_add_to_datetime_invalid_dt():
    initial_dt = "stringInput"
    datetime_offset = DatetimeOffset(unit="YS", magnitude=2)

    with pytest.raises(TypeError) as exc_info:
        datetime_offset.add_to_datetime(initial_dt)

    expected_error = (
        "Invalid initial datetime type: <class 'str'>. "
        "Expected type: cftime.datetime"
    )
    assert str(exc_info.value) == expected_error


def test_date_offset_unsupported_offset():
    with pytest.raises(ValueError) as exc_info:
        DatetimeOffset(unit="Y", magnitude=2)

    expected_error = (
        "Unsupported datetime offset: Y. "
        "Supported offsets: YS, MS, W, D, H, T, S"
    )
    assert str(exc_info.value) == expected_error


def test_parse_date_offset_no_offset_magnitude():
    with pytest.raises(ValueError) as exc_info:
        parse_date_offset("YS")

    expected_error = "No numerical value given for offset: YS"
    assert str(exc_info.value) == expected_error


@pytest.mark.parametrize(
        "start_date, end_date, caltype_int, expected",
        [
            (
                datetime.datetime(year=4, month=1, day=1),
                datetime.datetime(year=5, month=1, day=1),
                GREGORIAN,
                366 * SEC_PER_DAY
            ),
            (
                datetime.datetime(year=4, month=1, day=1),
                datetime.datetime(year=5, month=1, day=1),
                NOLEAP,
                365 * SEC_PER_DAY
            ),
            (
                datetime.datetime(year=300, month=1, day=1),
                datetime.datetime(year=301, month=1, day=1),
                GREGORIAN,
                365 * SEC_PER_DAY
            ),
            (
                datetime.datetime(year=400, month=1, day=1),
                datetime.datetime(year=400, month=12, day=31),
                GREGORIAN,
                365 * SEC_PER_DAY
            ),
            (
                datetime.datetime(year=12, month=7, day=22),
                datetime.datetime(year=23, month=3, day=15),
                GREGORIAN,
                (10 * 365 + 238) * SEC_PER_DAY
            ),
            (
                datetime.datetime(year=1, month=1, day=1),
                datetime.datetime(year=9999, month=1, day=1),
                GREGORIAN,
                (9998 * 365 + 2424) * SEC_PER_DAY
            )
        ]
)
def test_seconds_between_dates(start_date, end_date, caltype_int, expected):
    assert seconds_between_dates(start_date, end_date, caltype_int) == expected


@pytest.mark.parametrize(
        "date_int, expected",
        [
            (10101, datetime.date(1, 1, 1)),
            (100321, datetime.date(10, 3, 21)),
            (99991231, datetime.date(9999, 12, 31))
        ]
)
def test_int_to_date(date_int, expected):
    """
    Check that integers typically read in from namelists
    are correctly converted to datetime.date objects.
    """
    converted_date = int_to_date(date_int)
    assert converted_date == expected


@pytest.mark.parametrize(
        "bad_date_int",
        [0, 100000000, 101, -5, 11119153]
)
def test_int_to_date_failures(bad_date_int):
    """
    Check that int_to_date does not allow non existent
    or out of range dates.
    """
    with pytest.raises(ValueError):
        int_to_date(bad_date_int)


@pytest.mark.parametrize(
        "start_date, years, months, days, seconds, caltype, expected",
        [
            # Normal year
            (datetime.date(101, 1, 1), 1, 0, 0, 0, GREGORIAN, 365*SEC_PER_DAY),
            (datetime.date(101, 1, 1), 1, 0, 0, 0, NOLEAP, 365*SEC_PER_DAY),
            # Leap year
            (datetime.date(4, 1, 1), 1, 0, 0, 0, GREGORIAN, 366*SEC_PER_DAY),
            (datetime.date(4, 1, 1), 1, 0, 0, 0, NOLEAP, 365*SEC_PER_DAY),
            # Non-leap year due to 100 year rule
            (datetime.date(100, 1, 1), 1, 0, 0, 0, GREGORIAN, 365*SEC_PER_DAY),
            (datetime.date(100, 1, 1), 1, 0, 0, 0, NOLEAP, 365*SEC_PER_DAY),
            # Leap year due to 400 year rule
            (datetime.date(400, 1, 1), 1, 0, 0, 0, GREGORIAN, 366*SEC_PER_DAY),
            (datetime.date(500, 1, 1), 1, 0, 0, 0, NOLEAP, 365*SEC_PER_DAY),
            # February in leap years
            (datetime.date(40, 2, 8), 0, 1, 0, 0, GREGORIAN, 29*SEC_PER_DAY),
            (datetime.date(40, 2, 8), 0, 1, 0, 0, NOLEAP, 28*SEC_PER_DAY),
            # Misc
            (datetime.date(1, 1, 1), 0, 0, 0, 86400,
             GREGORIAN, 86400),
            # Max & min limits
            (datetime.date(1, 1, 1), 9998, 11, 30, 0,
             NOLEAP, (9998 * 365 + 364) * SEC_PER_DAY),
            (datetime.date(1, 1, 1), 9998, 11, 30, 0,
             GREGORIAN, (9998 * 365 + 2424 + 364) * SEC_PER_DAY),
            (datetime.date(1, 1, 1), 0, 0, 0, 1,
             GREGORIAN, 1),
            (datetime.date(1, 1, 1), 0, 0, 0, 1,
             NOLEAP, 1),
        ]
)
def test_runtime_from_date(
        start_date,
        years,
        months,
        days,
        seconds,
        caltype,
        expected):
    """
    Test that the number of seconds calculated for run lengths is correct.
    """
    runtime = runtime_from_date(start_date,
                                years,
                                months,
                                days,
                                seconds,
                                caltype)

    assert runtime == expected



"""Run the `payu checkout` command.

:copyright: Copyright 2018 Marshall Ward, see AUTHORS for details.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from pathlib import Path

from payu.branch import checkout_branch
import payu.subcommands.args as args

title = 'checkout'
parameters = {'description': ('A wrapper around git checkout. '
                              'Create a new branch (if specified), '
                              'checkout branch, setup experiment metadata '
                              'and create/switch archive/work symlinks')}

arguments = [args.model, args.config, args.laboratory, args.new_branch,
             args.branch_name, args.start_point, args.restart_path,
             args.keep_uuid, args.parent_experiment]


def transform_strings_to_path(path_str=None):
    return Path(path_str) if path_str is not None else None


def runcmd(model_type, config_path, lab_path, new_branch,
           branch_name, start_point,
           restart_path, keep_uuid, parent_experiment):
    """Execute the command."""
    config_path = transform_strings_to_path(config_path)
    lab_path = transform_strings_to_path(lab_path)
    restart_path = transform_strings_to_path(restart_path)

    checkout_branch(is_new_branch=new_branch,
                    branch_name=branch_name,
                    start_point=start_point,
                    restart_path=restart_path,
                    config_path=config_path,
                    lab_path=lab_path,
                    model_type=model_type,
                    keep_uuid=keep_uuid,
                    parent_experiment=parent_experiment)


runscript = runcmd
